"""
Created on Wed Nov 13 16:32:05 2024

Author: Sander van Nielen
"""
import json

import pandas as pd
from _mapping_type import add_mapping_comment
from _unique_id import add_unique_identifiers


def read_showvoc(json_file, short_names: list, path="."):
    """Read a JSON-LD file (as downloaded from https://showvoc.op.europa.eu/)
    and save it as a table in CSV format
    """
    short_names = [name.lower() for name in short_names]
    names = ["tree_" + name for name in short_names]
    data = {name: [] for name in names}

    with open(json_file, "r") as file:
        raw_json = json.load(file)
    # Navigate through the JSON structure
    for item in raw_json:
        for key in item:
            if key != "@id":
                for match_ in item[key]:
                    # Store the code of the source classification and its matching target
                    data[names[0]].append(item["@id"].split("/")[-1])
                    data[names[1]].append(match_["@id"].split("/")[-1])

    # Create a dataframe, including some empty columns
    df = pd.DataFrame(data)
    for new_col in ["comment", "skos_uri"]:
        df[new_col] = None

    # Save as CSV, and apply filler functions to the empty columns
    csv_path = f"{path}/conc_{short_names[0]}_{short_names[1]}.csv"
    df.to_csv(csv_path, index=False)
    add_mapping_comment(csv_path)
    add_unique_identifiers(csv_path)


json_file = "C:\\Users\\nielenssvan\\OneDrive - Universiteit Leiden\\BONSAI\\correspondence_tables\\CPA_2_1_CPC_2_1.jsonld"
read_showvoc(json_file, short_names=["CPA_2_1", "CPC_2_1"])



from classifications._pint_units import get_unit_registry


def test_units():
    def almost_equal(a: float, b: float, tol: float = float("1e-6")) -> bool:
        return abs(a - b) <= tol

    ureg = get_unit_registry()

    a = 1000000 * ureg.DKK_2017
    a_result = a.to("Meuro_2016")

    b = 1 * ureg.t
    b_result = b.to("tonnes")

    c = 1 * ureg.kWh
    c_result = c.to("TJ")

    assert almost_equal(a_result.m, 0.1416874676815668658358755478) == True
    assert a_result.u == "Meuro_2016"
    assert almost_equal(b_result.m, 1.0) == True
    assert b_result.u == "tonnes"
    assert almost_equal(c_result.m, 0.000003600) == True
    assert c_result.u == "terajoule"



# -*- coding: utf-8 -*-
"""
Created on Wed May  7 13:37:04 2025

@author: BertramdeBoer

This script is used to obtain historic and current mapping of countries to currencies,
according to ISO4217 obtained from:
https://www.iso.org/iso-4217-currency-codes.html

ISO4217 is maintained by:
https://www.six-group.com/en/products-services/financial-information/data-standards.html

Comments:
2025-05-09 BB: The withdrawal dates for historical currencies in the original
            source file seem to be inconsistent. The wikipedia page on this standard
            seems to have more robust data:
            https://en.wikipedia.org/wiki/ISO_4217#Historical_codes
            However, the data from the wikipedia page does not refer a source,
            so it is not certain whether these data are correct.

2025-05-09 BB: CountryConverter does not seem to contain former countries.
"""

import country_converter as coco
import pandas as pd
import requests
import xml.etree.ElementTree as ET


def get_publication_date(xml):
    """
    Extracts the publication date from the XML content.

    Parameters:
    xml (str): XML content as a string.

    Returns:
    str: Publication date extracted from the XML.
    """
    # Parse the XML content with ElementTree
    root = ET.fromstring(xml)

    # Extract the Pblshd attribute
    return root.attrib["Pblshd"]


def cntr_name2iso3(row):
    """
    Converts country name to ISO3 code.

    Parameters:
    row (pd.Series): A row of the DataFrame containing country names.

    Returns:
    str: ISO3 country code.
    """

    cntr_name = row["CtryNm"]
    cntr_name = cntr_name.replace("(THE)", "")  #'CONGO (THE)' not found by coco.
    cntr_name = cntr_name.rstrip()  # remove trailing space after removing '(THE)'
    return coco.convert(cntr_name)


def currency_name_rstrip(row):
    """
    Removes trailing spaces from currency name.

    Parameters:
    row (pd.Series): A row of the DataFrame containing currency names.

    Returns:
    str: currency name without trailing space.
    """
    return row["CcyNm"].rstrip()


def split_withdrawal_date(row):
    """
    Splits the withdrawal date into year and month components.

    Parameters:
    row (pd.Series): A row of the DataFrame containing withdrawal dates.

    Returns:
    tuple: Year start, month start, year end, month end.
    """

    s_withdrawal_date = row["WthdrwlDt"]
    s_withdrawal_date_split = s_withdrawal_date.split("-")

    yr_start_default = -1
    m_start_default = -1
    m_end_default = -1
    if len(s_withdrawal_date_split) == 2:
        if len(s_withdrawal_date_split[1]) == 4:  # year-year
            yr_start, yr_end = s_withdrawal_date_split
            m_start = m_start_default
            m_end = m_end_default
        else:
            yr_end, m_end = s_withdrawal_date_split  # year-month
            yr_start = yr_start_default
            m_start = m_start_default

    else:
        s_withdrawal_date_split = s_withdrawal_date.split(" to ")
        if (
            len(s_withdrawal_date_split[0]) == 4
            and len(s_withdrawal_date_split[1]) == 4
        ):  # year to year
            yr_start, yr_end = s_withdrawal_date_split
            m_start = m_start_default
            m_end = m_end_default
        else:  # year-month to year-month
            yr_start, m_start = s_withdrawal_date_split[0].split("-")
            yr_end, m_end = s_withdrawal_date_split[1].split("-")

    return int(yr_start), int(m_start), int(yr_end), int(m_end)


def yr_m2date(row):
    """
    Converts year and month components to datetime.

    Parameters:
    row (pd.Series): A row of the DataFrame containing year and month components.

    Returns:
    tuple: Start date and end date as datetime objects.
    """

    # if no start date is given, assume Unix start time: 01-jan-1970.
    if row["yr_start"] != -1:
        yr_start = row["yr_start"]
    else:
        yr_start = 1970
    if row["m_start"] != -1:
        m_start = row["m_start"]
    else:
        m_start = 1

    yr_end = row["yr_end"]

    # if no end month is given, assume December.
    if row["m_end"] != -1:
        m_end = row["m_end"]
    else:
        m_end = 12
    return pd.to_datetime(f"{yr_start}-{m_start}", format="%Y-%m"), pd.to_datetime(
        f"{yr_end}-{m_end}", format="%Y-%m"
    )


def fill_history_start_date_end_date(df_history):
    """
    Fills missing start dates with end dates from previous entries.

    Parameters:
    df_history (pd.DataFrame): DataFrame containing historical currency data.

    Returns:
    pd.DataFrame: DataFrame with filled start and end dates.
    """

    # Parse start and end dates.
    df_history[["yr_start", "m_start", "yr_end", "m_end"]] = df_history.apply(
        split_withdrawal_date, axis=1, result_type="expand"
    )
    df_history[["start_date_iso4217", "end_date_iso4217"]] = df_history.apply(
        yr_m2date, axis=1, result_type="expand"
    )

    # l_exclude = ["not found"]

    date_default = pd.to_datetime("1970-01")

    cntr_name_col = "CtryNm"
    df_history_cntr_uniq = df_history[cntr_name_col].unique()
    l_history_cntr_name_sort = []
    # For each country
    for cntr_name in df_history_cntr_uniq:
        # if cntr_name not in l_exclude:
        df_history_cntr_name = df_history[df_history[cntr_name_col] == cntr_name].copy()
        df_history_cntr_name_sort = df_history_cntr_name.sort_values("end_date_iso4217")
        # For each entry, if start date missing, fill with previous end date.
        for row_id, t_row in enumerate(df_history_cntr_name_sort.iterrows()):
            row_idx, row = t_row
            if not row_id:
                start_date = row["start_date_iso4217"]
                end_date = row["end_date_iso4217"]
            else:
                start_date = row["start_date_iso4217"]
                if start_date == date_default:
                    df_history_cntr_name_sort.loc[row_idx, "start_date_iso4217"] = (
                        end_date
                    )
                end_date = row["end_date_iso4217"]
        # Append country DataFrame
        l_history_cntr_name_sort.append(df_history_cntr_name_sort)
    df_history_cntr_name_sort_concat = pd.concat(l_history_cntr_name_sort)
    df_history_cntr_name_sort_concat = df_history_cntr_name_sort_concat.drop(
        ["yr_start", "m_start", "yr_end", "m_end"], axis=1
    )
    return df_history_cntr_name_sort_concat


def fill_current_start_date_end_date(df_current, df_history_date_fill, date_current):
    """
    Fills missing start dates in the current DataFrame using historical data.

    Parameters:
    df_current (pd.DataFrame): DataFrame containing current currency data.
    df_history_date_fill (pd.DataFrame): DataFrame with filled historical data.

    Returns:
    pd.DataFrame: DataFrame with filled start and end dates.
    """

    date_default = pd.to_datetime("1970-01")

    for row_id, row in df_current.iterrows():
        cntr_iso3 = row["cntr_iso3"]
        # if cntr in history, take end date as start date.
        if cntr_iso3 in df_history_date_fill["cntr_iso3"].values:
            df_history_date_fill_cntr_iso3 = df_history_date_fill[
                df_history_date_fill["cntr_iso3"] == cntr_iso3
            ]
            df_current.loc[row_id, "start_date_iso4217"] = max(
                df_history_date_fill_cntr_iso3["end_date_iso4217"]
            )
            df_current.loc[row_id, "start_date_iso4217"] = max(
                df_history_date_fill_cntr_iso3["end_date_iso4217"]
            )
        # else, start date is unix start date.
        else:
            df_current.loc[row_id, "start_date_iso4217"] = date_default

        df_current.loc[row_id, "end_date_iso4217"] = pd.to_datetime(date_current)

    return df_current


def fill_date(df_current, df_history, date_current):
    """
    Fills missing dates in both current and historical DataFrames and concatenates them.

    Parameters:
    df_current (pd.DataFrame): DataFrame containing current currency data.
    df_history (pd.DataFrame): DataFrame containing historical currency data.

    Returns:
    pd.DataFrame: Concatenated DataFrame with filled dates.
    """

    # Fill missing start dates with end dates from previous entry.
    df_history_date_fill = fill_history_start_date_end_date(df_history)
    df_current_date_fill = fill_current_start_date_end_date(
        df_current, df_history_date_fill, date_current
    )

    df_date_fill = pd.concat([df_current_date_fill, df_history_date_fill])

    # Sort first by country, and then by end date of currency.
    df_date_fill = df_date_fill.sort_values(["CtryNm", "end_date_iso4217"])
    df_date_fill = df_date_fill.reset_index(drop=True)

    return df_date_fill


def get_iso4217_wiki_history():
    """
    Get historical ISO 4217 currency data from Wikipedia.

    Returns:
    pd.DataFrame: The second DataFrame from the parsed HTML,
    containing historical currency data.
    """

    # Parse the ISO4271 wikipedia page.
    l_iso4217_wiki = pd.read_html("https://en.wikipedia.org/wiki/ISO_4217")

    # Historic data is the second DataFrame.
    return l_iso4217_wiki[2]


def wiki_remove_footnote(val):
    """
    Remove footnote references, e.g. [a], from values in wikipedia table.

    Parameters:
    val (str or any): The value from which to remove the footnote reference.

    Returns:
    str or any: The value without the footnote reference if it was a string,
    otherwise the original value.
    """

    if type(val) == str:
        return val.split("[")[0]
    else:
        return val


def wiki_to_datetime(row):
    """
    Convert 'From' and 'Until' columns from wikipedia table to datetime.

    Note that some cells have {{dunno}}, which are not machine-readable, but valid:
    https://en.wikipedia.org/wiki/Template:Dunno

    2025-05-09 BB: I made a minor revision to the wikipedia page, to improve the
    machine-readability by changing a hyphen from U+2013 to U+002D
    https://en.wikipedia.org/w/index.php?title=ISO_4217&oldid=1289566835

    Parameters:
    row (pd.Series): A row of the DataFrame.

    Returns:
    tuple: A tuple containing the converted 'From' and 'Until' dates as
    datetime objects.
    """

    return pd.to_datetime(row["From"], errors="coerce"), pd.to_datetime(
        row["Until"], errors="coerce"
    )


if __name__ == "__main__":

    save_raw = True
    save_proc = True
    save_bonsai_tree = True
    save_bonsai_conc = True

    # URLs for current and history ISO 4217 country-currency mapping
    url_current = (
        "https://www.six-group.com/dam/download/financial-information/"
        "data-center/iso-currrency/lists/list-one.xml"
    )
    url_history = (
        "https://www.six-group.com/dam/download/financial-information/"
        "data-center/iso-currrency/lists/list-three.xml"
    )

    # Get XMLs for country-currency mappings
    xml_current = requests.get(url_current).content
    xml_history = requests.get(url_history).content

    # Cast XMLs to pandas DataFrames.
    df_current = pd.read_xml(xml_current, xpath=".//CcyNtry")
    df_history = pd.read_xml(xml_history, xpath=".//HstrcCcyNtry")

    df_current_raw = df_current.copy()
    df_history_raw = df_history.copy()

    # Get XML publication dates.
    date_current = get_publication_date(xml_current)
    date_history = get_publication_date(xml_history)

    # Save raw ISO4217 to csv.
    if save_raw:
        df_current.to_csv(f"iso4217_current_raw_{date_current}.csv", index=False)
        df_history.to_csv(f"iso4217_history_raw_{date_history}.csv", index=False)

    # Remove trailing spaces in currency names.
    df_current["CcyNm"] = df_current.apply(currency_name_rstrip, axis=1)
    df_history["CcyNm"] = df_history.apply(currency_name_rstrip, axis=1)

    # Add ISO3 country codes.
    df_current["cntr_iso3"] = df_current.apply(cntr_name2iso3, axis=1)
    df_history["cntr_iso3"] = df_history.apply(cntr_name2iso3, axis=1)

    # Fill start and end dates, and concatenate history and current ISO4217.
    df_iso4217_date = fill_date(df_current, df_history, date_current)

    # Get historical ISO 4217 currency data from Wikipedia.
    df_iso4217_wiki_history = get_iso4217_wiki_history()

    # Remove footnotes, e.g. [a], from values.
    pd_major, pd_minor, pd_patch = pd.__version__.split(".")
    if int(pd_major) >= 2 and int(pd_minor) >= 1:
        df_iso4217_wiki_history = df_iso4217_wiki_history.map(wiki_remove_footnote)
    else:
        df_iso4217_wiki_history = df_iso4217_wiki_history.applymap(wiki_remove_footnote)

    # Cast from and until dates to datetime.
    df_iso4217_wiki_history[["From", "Until"]] = df_iso4217_wiki_history.apply(
        wiki_to_datetime, axis=1, result_type="expand"
    )

    # Merge from and until dates with iso4217 DataFrame.
    df_iso4217_date_wiki = pd.merge(
        df_iso4217_date,
        df_iso4217_wiki_history[["Code", "From", "Until"]],
        left_on="Ccy",
        right_on="Code",
        how="left",
    )
    df_iso4217_date_wiki = df_iso4217_date_wiki.drop(columns=["Code"])
    df_iso4217_date_wiki = df_iso4217_date_wiki.rename(
        {"From": "start_date_wiki", "Until": "end_date_wiki"}, axis=1
    )

    # Drop US Dollar Same Day and Next Day, change USD start date to Unix start.
    df_iso4217_date_wiki = df_iso4217_date_wiki[df_iso4217_date_wiki["Ccy"] != "USS"]
    df_iso4217_date_wiki = df_iso4217_date_wiki[df_iso4217_date_wiki["Ccy"] != "USN"]
    df_iso4217_date_wiki.loc[
        df_iso4217_date_wiki["Ccy"] == "USD", "start_date_iso4217"
    ] = "1970-01-01"

    if save_proc:
        df_iso4217_date_wiki.to_csv(f"iso4217_current_history_wiki_{date_current}.csv")

    if save_bonsai_tree:
        df_tree = df_iso4217_date_wiki[["Ccy", "CcyNm", "CcyNbr"]]
        df_tree = df_tree.drop_duplicates("Ccy")
        df_tree = df_tree.sort_values("CcyNbr")
        df_tree = df_tree.reset_index(drop=True)
        df_tree = df_tree.rename(
            {"Ccy": "code", "CcyNm": "name", "CcyNbr": "alias_code"}, axis=1
        )
        df_tree["description"] = ""
        df_tree["comment"] = ""
        df_tree.to_csv("dim_bonsai.csv", float_format="%.0f", index=False)

    if save_bonsai_conc:
        df_conc = df_iso4217_date_wiki[
            [
                "Ccy",
                "cntr_iso3",
                "start_date_iso4217",
                "end_date_iso4217",
                "start_date_wiki",
                "end_date_wiki",
            ]
        ]
        df_conc = df_conc.rename({"Ccy": "currency_code_alpha"}, axis=1)
        df_conc.to_csv("bonsai_currency_to_country.csv", index=False)



"""Script to explore task flag of package classifications

Use this folder for any work required to identify content of
data/flag/

This template illustrates identification of entity-relation model
"""
from logging import getLogger

import numpy as np
import pandas as pd
from dataio import datapackage as iodp
from templates import set_logger

logger = getLogger("root")
set_logger(
    filename=("classifications_flag.log"),
    path=".",
    log_level=20,
    overwrite=True,
    create_path=True,
)


logger.info("Started script to generate clean_classifications_flag.dataio.yam")

metadata_file = "clean_classifications_flag.dataio.yaml"
iodp.describe(full_path=metadata_file, overwrite=True)
iodp.plot(full_path=metadata_file, overwrite=True)
iodp.validate(full_path=metadata_file, overwrite=True)

logger.info("Finished script to generate clean_classifications_flag.dataio.yam")



import os
from pathlib import Path

import pandas as pd
import pytest
import yaml

import classifications


def test_lookup():
    result = classifications.location.datapackage.dim_level.lookup("layer")

    assert len(result.index) == 7


def test_get_children_failed():
    with pytest.raises(KeyError):
        classifications.location.datapackage.dim_level.get_children("layer")


def test_get_children():
    result = classifications.activitytype.datapackage.tree_nace_rev2.get_children(
        "01.3"
    )

    assert len(result.index) == 1


def test_get_children_multiple():
    result = classifications.activitytype.datapackage.tree_nace_rev2.get_children(
        ["01.3", "01.2"]
    )

    assert len(result.index) == 10


def test_get_children_deep():
    result = classifications.activitytype.datapackage.tree_nace_rev2.get_children(
        ["01", "03"]
    )

    assert len(result.index) == 38 + 6


def test_get_children_flat():
    result = classifications.activitytype.datapackage.tree_nace_rev2.get_children(
        ["01", "03"], deep=False
    )

    assert len(result.index) == 7 + 2


def test_get_children_with_parent():
    result = classifications.activitytype.datapackage.tree_nace_rev2.get_children(
        "01.3", return_parent=True
    )

    assert len(result.index) == 2


def test_get_children_with_parent_multiple():
    result = classifications.activitytype.datapackage.tree_nace_rev2.get_children(
        ["01.3", "01.2"], return_parent=True
    )

    assert len(result.index) == 12


def test_create_conc():
    A = pd.DataFrame(  # flowobject_from,flowobject_to,classification_from,classification_to
        {
            "activitytype_from": [
                "A_a",
                "A_b",
                "A_c",
                "A_c",
                "A_d",
                "A_d",
                "A_e",
                "A_f",
                "A_x",
            ],
            "activitytype_to": ["1", "1", "2", "3", "5", "6", "7", "7", "x"],
            "classification_from": 9 * ["XXX"],
            "classification_to": 9 * ["YYY"],
        }
    )
    B = pd.DataFrame(
        {
            "activitytype_from": [
                "111",
                "222",
                "444",
                "444",
                "555",
                "666",
                "777",
                "888",
            ],
            "activitytype_to": ["1", "1", "2", "3", "4", "5", "6", "7"],
            "classification_from": 8 * ["ZZZ"],
            "classification_to": 8 * ["YYY"],
        }
    )
    df_result = classifications.create_conc(A, B, source="XXX", target="YYY")

    df_expected = pd.DataFrame(
        {
            "activitytype_from": [
                "A_a",
                "A_a",
                "A_b",
                "A_b",
                "A_c",
                "A_d",
                "A_d",
                "A_e",
                "A_f",
            ],
            "activitytype_to": [
                "111",
                "222",
                "111",
                "222",
                "444",
                "666",
                "777",
                "888",
                "888",
            ],
            "classification_from": 9 * ["XXX"],
            "classification_to": 9 * ["YYY"],
            "comment": [
                "many-to-many correspondence",
                "many-to-many correspondence",
                "many-to-many correspondence",
                "many-to-many correspondence",
                "one-to-one correspondence",
                "one-to-many correspondence",
                "one-to-many correspondence",
                "many-to-one correspondence",
                "many-to-one correspondence",
            ],
            "skos_uri": [
                "http://www.w3.org/2004/02/skos/core#relatedMatch",
                "http://www.w3.org/2004/02/skos/core#relatedMatch",
                "http://www.w3.org/2004/02/skos/core#relatedMatch",
                "http://www.w3.org/2004/02/skos/core#relatedMatch",
                "http://www.w3.org/2004/02/skos/core#exactMatch",
                "http://www.w3.org/2004/02/skos/core#narrowMatch",
                "http://www.w3.org/2004/02/skos/core#narrowMatch",
                "http://www.w3.org/2004/02/skos/core#broadMatch",
                "http://www.w3.org/2004/02/skos/core#broadMatch",
            ],
        }
    )
    pd.testing.assert_frame_equal(df_expected, df_result)


def test_disaggregate_tree_bonsai_fails():
    with pytest.raises(ValueError):
        classifications.activitytype.datapackage.disaggregate_bonsai(
            codes_or_yaml={
                "disaggregations": [{"old_code": "ai", "new_codes": [{"code": "AI"}]}]
            }
        )


def test_disaggregate_tree_bonsai_pass():
    yaml_content = """
    disaggregations:
      - old_code: "ai"
        new_codes:
          - code: "AI"
            description: "scacaec"
            mappings: {}
          - code: "AAASSSS"
            description: ""
            mappings: {}
    """
    d = classifications.activitytype.datapackage.disaggregate_bonsai(
        codes_or_yaml=yaml.safe_load(yaml_content)
    )

    assert d["tree_bonsai"].iloc[-1]["code"] == "AAASSSS"
    assert d["tree_bonsai"].iloc[-1]["name"] == ""
    assert d["tree_bonsai"].iloc[-2]["code"] == "AI"
    assert d["tree_bonsai"].iloc[-2]["name"] == "scacaec"


def test_disaggregate_conc_bonsai_cpc_2_1():
    yaml_content = """
    disaggregations:
      - old_code: "fi_01929_06"
        new_codes:
          - code: "C_Agavs_0"
            description: "scacaec"
            mappings:
              cpc_2_1:
                - "1929"
          - code: "C_Agavs_0"
            description: ""
            mappings: {}
    """
    d = classifications.flowobject.datapackage.disaggregate_bonsai(
        codes_or_yaml=yaml.safe_load(yaml_content)
    )

    assert d["conc_bonsai_cpc_2_1"].iloc[-1]["flowobject_from"] == "C_Agavs_0"
    assert d["conc_bonsai_cpc_2_1"].iloc[-1]["flowobject_to"] == "1929"

    yaml_content = """
    disaggregations:
      - old_code: "fi_01929_06"
        new_codes:
          - code: "C_Agavs_0"
            description: "scacaec"
            mappings:
              cpc_2_1:
                - "1929"
          - code: "C_Agavs_1"
            description: ""
            mappings:
              cpc_2_1:
                - "1929"
    """
    d = classifications.flowobject.datapackage.disaggregate_bonsai(
        codes_or_yaml=yaml.safe_load(yaml_content)
    )

    assert d["conc_bonsai_cpc_2_1"].iloc[-1]["flowobject_from"] == "C_Agavs_1"
    assert d["conc_bonsai_cpc_2_1"].iloc[-1]["flowobject_to"] == "1929"
    assert d["conc_bonsai_cpc_2_1"].iloc[-2]["flowobject_from"] == "C_Agavs_0"
    assert d["conc_bonsai_cpc_2_1"].iloc[-2]["flowobject_to"] == "1929"
    assert (
        "Agavs" in d["conc_bonsai_cpc_2_1"]["flowobject_from"].values
    ) == False  # check that old code is removed


def test_disaggregate_tree_bonsai_pass_yaml():
    test_path = str(
        Path(os.path.dirname(__file__)).parent / "tests/data/disaggregate.yaml"
    )

    d = classifications.activitytype.datapackage.disaggregate_bonsai(
        codes_or_yaml=test_path
    )

    assert d["tree_bonsai"].iloc[-1]["code"] == "AAASSSS"
    assert d["tree_bonsai"].iloc[-1]["name"] == ""
    assert d["tree_bonsai"].iloc[-2]["code"] == "AI"
    assert d["tree_bonsai"].iloc[-2]["name"] == "scacaec"


def test_disaggragate_multiple():
    codes = {
        "disaggregations": [
            {
                "old_code": "fi_33421",
                "new_codes": [
                    {
                        "code": "C_ETHYLENE",
                        "description": "production of ethylene",
                        "mappings": {"prodcom_total_2_0": ["20141130"]},
                    },
                    {
                        "code": "C_CHEMrest",
                        "description": "production of rest",
                        "mappings": {
                            "prodcom_total_2_0": ["20141140", "20141150", "20141160"]
                        },
                    },
                ],
            }
        ]
    }
    d = classifications.flowobject.datapackage.disaggregate_bonsai(codes_or_yaml=codes)
    assert d["tree_bonsai"].iloc[-1]["code"] == "C_CHEMrest"
    assert d["tree_bonsai"].iloc[-1]["level"] == 6
    assert d["tree_bonsai"].iloc[-2]["code"] == "C_ETHYLENE"
    assert d["tree_bonsai"].iloc[-2]["level"] == 6
    assert (
        d["conc_bonsai_prodcom_total_2_0"].iloc[-1]["flowobject_from"] == "C_CHEMrest"
    )
    assert d["conc_bonsai_prodcom_total_2_0"].iloc[-1]["flowobject_to"] == "20141160"
    assert (
        d["conc_bonsai_prodcom_total_2_0"].iloc[-2]["flowobject_from"] == "C_CHEMrest"
    )
    assert d["conc_bonsai_prodcom_total_2_0"].iloc[-2]["flowobject_to"] == "20141150"
    assert (
        d["conc_bonsai_prodcom_total_2_0"].iloc[-3]["flowobject_from"] == "C_CHEMrest"
    )
    assert d["conc_bonsai_prodcom_total_2_0"].iloc[-3]["flowobject_to"] == "20141140"
    assert (
        d["conc_bonsai_prodcom_total_2_0"].iloc[-4]["flowobject_from"] == "C_ETHYLENE"
    )
    assert d["conc_bonsai_prodcom_total_2_0"].iloc[-4]["flowobject_to"] == "20141130"


def test_convert_name():
    assert (
        classifications.location.datapackage.conc_regex_bonsai.convert_name(
            "Bosnia and Herzegovina"
        )
        == "BA"
    )
    assert (
        classifications.location.datapackage.conc_regex_bonsai.convert_name(
            "Congo (Brazzaville)"
        )
        == "CG"
    )
    assert classifications.location.datapackage.conc_regex_bonsai.convert_name(
        "Serbia and Montenegro"
    ) == ["ME", "RS"]



# generate_excel.py
import os

import pandas as pd

# def generate_excel_files():
#    # Get the directory where the script is located
#    script_dir = os.path.dirname(os.path.abspath(__file__))
#
#    directories = [
#        "flow/activitytype",
#        # "dataquality",
#        "flow/flowobject",
#        # "location",
#        # "time",
#        # "uncertainty",
#    ]
#
#    # Define the output directory for Excel files within the Sphinx _static directory
#    excel_dir = os.path.join(script_dir, "_static", "excel_files")
#
#    # Create the output directory if it doesn't exist
#    os.makedirs(excel_dir, exist_ok=True)
#
#    # Convert all CSV files to Excel
#    for d in directories:
#        csv_dir = os.path.normpath(
#            os.path.join(script_dir, "..", "src", "classifications", "data", d)
#        )
#
#        dict_with_dfs = {}
#
#        for csv_file in os.listdir(csv_dir):
#            if csv_file.endswith(".csv") and "conc_bonsai" in csv_file and "bonsai.csv" in csv_file:
#                # Read the CSV file
#                csv_path = os.path.join(csv_dir, csv_file)
#                df = pd.read_csv(csv_path, encoding="utf-8", sep=",", dtype="str")
#                dict_with_dfs[f"{csv_file}"] = df
#            if csv_file.endswith(".csv") and "tree_bonsai" in csv_file:
#                # Read the CSV file
#                csv_path = os.path.join(csv_dir, csv_file)
#                df_merged = pd.read_csv(
#                    csv_path, encoding="utf-8", sep=",", dtype="str"
#                )
#
#        matching_keys = [key for key in dict_with_dfs if "tree_bonsai" in key]
#        if len(matching_keys) > 1:
#            raise ValueError(
#                f"Too many files with 'tree_bonsai' in filename inside {d}"
#            )
#        try:
#            for k, df in dict_with_dfs.items():
#                if k != "tree_bonsai":
#                    if "activitytype_from" in df.columns:
#                        if df["classification_from"][0].lower() == "bonsai":
#                             column_to_merge = "activitytype_from"
#                             class_name = df["classification_to"][0]
#                             df = df.drop(columns=["comment", "skos_uri", "alias_code"])
#                             class_name = df["classification_to"][0]
#                             df = df.rename(columns={"activitytype_to": class_name})
#                             df = (
#                                 df.groupby("activitytype_from")[df.columns[1]]
#                                 .apply(list)
#                                 .reset_index()
#                             )
#
#                             df_merged = pd.merge(
#                                 df_merged,
#                                 df,
#                                 left_on="code",
#                                 right_on=str(column_to_merge),
#                                 how="left",
#                             )
#                        elif df["classification_to"][0].lower() == "bonsai":
#                            column_to_merge = "activitytype_to"
#                            class_name = df["classification_from"][0]
#                            df = df.drop(columns=["comment", "skos_uri", "alias_code"], errors="ignore")
#                            df = df.rename(columns={"activitytype_from": class_name})
#                            df = (
#                                df.groupby(column_to_merge)[df.columns[1]]
#                                .apply(list)
#                                .reset_index()
#                            )
#                            df_merged = pd.merge(
#                                df_merged,
#                                df,
#                                left_on="code",
#                                right_on=column_to_merge,
#                                how="left",
#                            )
#                    if "flowobject_from" in df.columns:
#                        if df["classification_from"][0].lower() == "bonsai":
#                             column_to_merge = "flowobject_from"
#                             class_name = df["classification_to"][0]
#                             df = df.drop(columns=["comment", "skos_uri", "alias_code"])
#                             class_name = df["classification_to"][0]
#                             df = df.rename(columns={"flowobject_to": class_name})
#                             df = (
#                                 df.groupby("flowobject_from")[df.columns[1]]
#                                 .apply(list)
#                                 .reset_index()
#                             )
#
#                             df_merged = pd.merge(
#                                 df_merged,
#                                 df,
#                                 left_on="code",
#                                 right_on=str(column_to_merge),
#                                 how="left",
#                             )
#                        elif df["classification_to"][0].lower() == "bonsai":
#                            column_to_merge = "flowobject_to"
#                            class_name = df["classification_from"][0]
#                            df = df.drop(columns=["comment", "skos_uri", "alias_code"], errors="ignore")
#                            df = df.rename(columns={"flowobject_from": class_name})
#                            df = (
#                                df.groupby(column_to_merge)[df.columns[1]]
#                                .apply(list)
#                                .reset_index()
#                            )
#                            df_merged = pd.merge(
#                                df_merged,
#                                df,
#                                left_on="code",
#                                right_on=column_to_merge,
#                                how="left",
#                            )
#            df_merged = df_merged.loc[
#                :, ~df_merged.columns.str.startswith("flowobject_from")
#            ]
#            df_merged = df_merged.loc[
#                :, ~df_merged.columns.str.startswith("activitytype_from")
#            ]
#
#        except (KeyError, IndexError):
#            pass
#
#        # Define the Excel file name and path
#        if d.startswith("flow/"):
#            d = d[len("flow/") :]
#        else:
#            pass
#        excel_file = f"{d}_tree_bonsai.xlsx"
#        excel_path = os.path.join(excel_dir, excel_file)
#        # Save the data as an Excel file
#        df_merged.to_excel(excel_path, index=False)
#
#
# if __name__ == "__main__":
#    generate_excel_files()


def merge_concordance(df_merged, df, entity_type):
    # Detect if bonsai is from or to
    classification_from = df["classification_from"].iloc[0].lower()
    classification_to = df["classification_to"].iloc[0].lower()

    if classification_from == "bonsai":
        from_col = f"{entity_type}_from"
        to_col = f"{entity_type}_to"
        class_name = df["classification_to"].iloc[0]
    elif classification_to == "bonsai":
        from_col = f"{entity_type}_to"
        to_col = f"{entity_type}_from"
        class_name = df["classification_from"].iloc[0]
    else:
        return df_merged  # Not a bonsai mapping

    # Clean and group
    df = df.drop(columns=["comment", "skos_uri", "alias_code"], errors="ignore")
    df = df.rename(columns={to_col: class_name})

    df_grouped = df.groupby(from_col)[class_name].apply(list).reset_index()

    df_merged = pd.merge(
        df_merged, df_grouped, left_on="code", right_on=from_col, how="left"
    )

    df_merged = df_merged.drop(columns=[from_col], errors="ignore")
    return df_merged


def generate_excel_files():
    script_dir = os.path.dirname(os.path.abspath(__file__))

    directories = [
        "flow/activitytype",
        "flow/flowobject",
    ]

    excel_dir = os.path.join(script_dir, "_static", "excel_files")
    os.makedirs(excel_dir, exist_ok=True)

    for d in directories:
        csv_dir = os.path.normpath(
            os.path.join(script_dir, "..", "src", "classifications", "data", d)
        )

        dict_with_dfs = {}
        df_merged = None

        for csv_file in os.listdir(csv_dir):
            csv_path = os.path.join(csv_dir, csv_file)

            # Load tree bonsai file
            if "tree_bonsai" in csv_file and csv_file.endswith(".csv"):
                df_merged = pd.read_csv(
                    csv_path, encoding="utf-8", sep=",", dtype="str"
                )

            # Load all conc_*.csv concordance files
            elif csv_file.startswith("conc_") and csv_file.endswith(".csv"):
                df = pd.read_csv(csv_path, encoding="utf-8", sep=",", dtype="str")

                # Check if bonsai is involved
                if (
                    df["classification_from"].iloc[0].lower() == "bonsai"
                    or df["classification_to"].iloc[0].lower() == "bonsai"
                ):
                    dict_with_dfs[csv_file] = df

        if df_merged is None:
            print(f"Warning: No tree_bonsai file found in {d}")
            continue

        for k, df in dict_with_dfs.items():
            if "activitytype_from" in df.columns or "activitytype_to" in df.columns:
                df_merged = merge_concordance(df_merged, df, "activitytype")
            elif "flowobject_from" in df.columns or "flowobject_to" in df.columns:
                df_merged = merge_concordance(df_merged, df, "flowobject")

        # Remove any leftover source columns
        df_merged = df_merged.loc[:, ~df_merged.columns.str.startswith("activitytype_")]
        df_merged = df_merged.loc[:, ~df_merged.columns.str.startswith("flowobject_")]

        # Save Excel
        d_clean = d.split("/")[-1]
        excel_file = f"{d_clean}_tree_bonsai.xlsx"
        excel_path = os.path.join(excel_dir, excel_file)
        df_merged.to_excel(excel_path, index=False)
        print(f"Saved: {excel_file}")


if __name__ == "__main__":
    generate_excel_files()



import importlib.resources
import re
from decimal import getcontext
from logging import getLogger

import pandas as pd
import pint
from pint.errors import DefinitionSyntaxError

logger = getLogger("root")

# Set Decimal precision
getcontext().prec = 28

start_year = 2011
end_year = 2025


def get_unit_registry():
    ureg = pint.UnitRegistry()

    ## MONETARY UNITS ##

    BASE_YEAR = 2020
    BASE_CURRENCY = f"EUR_{BASE_YEAR}"

    # --- Load exchange rates (USD base) ---
    with importlib.resources.path(
        "classifications.data.unit.monetary", "fact_currency_per_usd.csv"
    ) as exchange_rate_file:
        exchange_df = pd.read_csv(exchange_rate_file)

    # --- Load deflators for EUR ---
    with importlib.resources.path(
        "classifications.data.unit.monetary", "fact_deflator_deu.csv"
    ) as deflators_file:
        deflator_df = pd.read_csv(deflators_file)

    # --- Load dimensions for aliases ---
    with importlib.resources.path(
        "classifications.data.unit.monetary", "dim_bonsai.csv"
    ) as dim_file:
        dim_df = pd.read_csv(dim_file)

    # --- Define base EUR_<year> unit ---
    ureg.define(f"{BASE_CURRENCY} = [currency]")

    base_deflator = deflator_df.query("time == @BASE_YEAR")["value"].iloc[0]

    # 1 --- Define EUR_<year> from deflators ---
    for _, row in deflator_df.iterrows():
        year = int(row["time"])
        deflator = row["value"]
        factor = deflator / base_deflator
        unit_name = f"EUR_{year}"
        if unit_name != BASE_CURRENCY:
            definition = f"{unit_name} = {factor} * {BASE_CURRENCY}"
            ureg.define(definition)
            logger.debug(f"Defining Monetary Unit A: {definition}")

    # 2 --- Define USD_<year> in terms of EUR_<year> ---
    usd_rates = exchange_df[exchange_df["unit"] == "EUR/USD"]

    for _, row in usd_rates.iterrows():
        year = int(row["time"])
        rate_per_usd = row["value"]  # EUR per 1 USD
        eur_unit = f"EUR_{year}"
        usd_unit = f"USD_{year}"
        factor = 1.0 / rate_per_usd  # 1 USD = X EUR → 1 EUR = 1/X USD

        if eur_unit != usd_unit:  # Avoid self-reference
            definition = f"{usd_unit} = {factor} * {eur_unit}"
            ureg.define(definition)
            logger.debug(f"Defining Monetary Unit B: {definition}")

    # 3 --- Define all other currency_<year> in terms of USD_<year> ---
    other_rates = exchange_df[exchange_df["unit"] != "EUR/USD"]

    for _, row in other_rates.iterrows():
        year = int(row["time"])
        unit_str = row["unit"]
        rate_per_usd = row["value"]

        currency = unit_str.split("/")[0]  # e.g. "AFN/USD" → "AFN"
        unit_name = f"{currency}_{year}"
        usd_unit = f"USD_{year}"

        # Ensure we don’t define a unit in terms of itself
        if unit_name != usd_unit:
            factor = 1.0 / rate_per_usd
            try:
                definition = f"{unit_name} = {factor} * {usd_unit}"
                logger.debug(f"Defining Monetary Unit C: {definition}")
                ureg.define(definition)
            except Exception as e:
                logger.debug(f"Skipping {unit_name} due to error: {e}")

    # --- Define aliases ---
    alias_map = {
        str(row["code"]).strip(): str(row["alias"]).strip()
        for _, row in dim_df.iterrows()
    }

    for code, alias in alias_map.items():
        # Look for all units in registry that start with the code followed by an underscore and 4-digit year
        pattern = re.compile(rf"^{re.escape(code)}_(\d{{4}})$")

        for unit in list(ureg._units):  # Access defined unit names
            match = pattern.match(unit)
            if match:
                year = match.group(1)
                alias_unit = f"{alias}_{year}"
                try:
                    ureg.define(f"{alias_unit} = {unit}")
                    logger.debug(f"Defined alias: {alias_unit} = {unit}")
                except DefinitionSyntaxError:
                    logger.warning(
                        f"Warning: Could not define alias '{alias_unit}' = '{unit}' (syntax error)"
                    )
                except Exception as e:
                    logger.warning(
                        f"Warning: Could not define alias '{alias_unit}' = '{unit}': {e}"
                    )

    ## PHYSICAL UNITS ##
    with importlib.resources.path(
        "classifications.data.unit.physical", "dim_bonsai.csv"
    ) as pyhsical_unit_file:
        df2 = pd.read_csv(pyhsical_unit_file)

        definition = None

        for _, row in df2.iterrows():
            code = row["code"]
            name = row["name"]
            new_dim = row["new_pint_dimension"]
            scale = row["scale"]
            ref_unit = row["reference_unit"]
            alias = row["alias"]

            if pd.notna(new_dim) and new_dim.strip() != "":  # Define new base dimension
                dim_name = f"{new_dim}"
                if f"[{dim_name}]" not in ureg._dimensions:
                    logger.debug(f"Defining dimension: [{dim_name}]")
                    ureg.define(f"[{dim_name}]")
                definition = f"{code} = [{dim_name}]"
            else:  # Derived unit from reference
                ref_unit = f"{ref_unit}"
                scale_val = float(scale)
                definition = f"{code} = {scale_val} * {ref_unit}"

                # Add aliases
            if alias and not pd.isna(alias):
                definition += " = " + " = " + alias

            logger.debug(f"Defining Physical Unit: {definition}")
            ureg.define(definition)

    ## MAGNITUDES ##
    with importlib.resources.path(
        "classifications.data.unit.magnitude", "dim_bonsai.csv"
    ) as magnitude_unit_file:
        df3 = pd.read_csv(magnitude_unit_file)

        for _, row in df3.iterrows():
            code = row["code"]
            scale = row["scale"]

            ref_unit = f"{ref_unit}"
            scale_val = float(scale)
            definition = f"{code} = {scale_val}"

            logger.debug(f"Defining Magnitude: {definition}")
            ureg.define(definition)

    return ureg



# -*- coding: utf-8 -*-
"""
Created on Sun Jan 19 11:45:14 2025
A test function that checks the correspondence files of external codes to Bonsai
and the associated tree files. It checks for missing columns and missing values.

For convenience, at the bottom there is a function to update the column names
to match the new format.

Author: Sander van Nielen (AAU)
"""
from logging import getLogger
from pathlib import Path

import pandas as pd
import pytest

logger = getLogger("root")

# directories to be checked
activitytype_path = Path("src/classifications/data/flow/activitytype")
flowobject_path = Path("src/classifications/data/flow/flowobject")
flow_path = Path("src/classifications/data/flow")

# Function to load CSV and check required columns
def assert_columns(file: Path, required_columns: list):
    df = pd.read_csv(file, dtype=str)
    for col in required_columns:
        assert col in df.columns, f"'{col}' column missing in {file.name}"
    return df


@pytest.mark.parametrize(
    "directory",
    [activitytype_path, flowobject_path],
    ids=["activitytype", "flowobject"],
)
def test_tree_files(directory):
    test_pass = True

    # Check all tree_*.csv files
    tree_files = directory.glob("tree_*.csv")
    for tree_file in tree_files:
        df_tree = assert_columns(tree_file, ["level"])

        level_0 = len(df_tree[df_tree.level == "0"])
        if level_0 > 6:
            logger.warning(
                f"{level_0} items have level 0 in {directory.name}/{tree_file.name}"
            )

    # Check conc files
    conc_files = list(directory.glob("conc_*_bonsai.csv")) + list(
        directory.glob("conc_bonsai_*.csv")
    )
    for conc_file in conc_files:
        df_conc = assert_columns(
            conc_file,
            ["classification_from", "classification_to", "comment"],
        )
        # Check if the auto-filled columns are filled
        for col in ["comment"]:
            if df_conc[col].isna().any():
                logger.error(f"Empty '{col}' in {conc_file.name}")
                test_pass = False
        # Check for 'one-sided' links
        for col in [f"{directory.name}_from", f"{directory.name}_to"]:
            if df_conc[col].isna().any():
                logger.error(f"Empty values in '{col}' in {conc_file.name}")
                test_pass = False
                df_conc = df_conc[df_conc[col].notna()]

        #        # Search the associated tree file
        #        tree_filename = conc_file.name.replace("conc_", "tree_").replace("_bonsai", "")
        #        tree_file = directory / tree_filename
        #        if not tree_file.exists():
        #            logger.error(f"Missing file: {tree_file.name}")
        #            test_pass = False
        #            continue

        # Check the columns and levels of the tree file
        df_tree = assert_columns(tree_file, ["code", "name"])
        if "level" not in df_tree.columns:
            logger.error(f"'level' column missing in {tree_file.name}")
            test_pass = False
        else:
            level_0 = len(df_tree[df_tree.level == "0"])
            if level_0 > 5:
                logger.info(f"{level_0} items have level 0 in {tree_file.name}")

        #        # Check if source classification codes are in the associated tree file
        #        suffix = "_from" if "_bonsai.csv" in conc_file.name else "_to"
        #        col = directory.name + suffix
        #        missing_codes = set(df_conc[col]) - set(df_tree["code"])
        #        if missing_codes:
        #            logger.warning(
        #                f"Codes missing in {tree_file.name}: {missing_codes} (indicated by conc file: '{conc_file.name}')"
        #            )

        # Check if there are any 'many-to-many' entries (doesn't make the test fail)
        many_to_many = df_conc[df_conc.comment == "many-to-many correspondence"]
        if not many_to_many.empty:
            logger.warning(
                f"Many-to-many correspondences found in {conc_file.name}:\n"
                f"{many_to_many.iloc[:10, :2]}"
            )

    assert test_pass, "One or more file checks failed. See the warning messages."


@pytest.mark.parametrize(
    "directory",
    [activitytype_path, flowobject_path],
    ids=["activitytype", "flowobject"],
)
def test_conc_files(directory):
    # Check conc files
    conc_files = list(directory.glob("conc_*_bonsai.csv")) + list(
        directory.glob("conc_bonsai_*.csv")
    )
    for conc_file in conc_files:

        if "activitytype" in directory.name:
            df_conc = assert_columns(
                conc_file,
                [
                    "activitytype_from",
                    "activitytype_to",
                    "classification_from",
                    "classification_to",
                    "comment",
                ],
            )
        elif "flowobject" in directory.name:
            df_conc = assert_columns(
                conc_file,
                [
                    "flowobject_from",
                    "flowobject_to",
                    "classification_from",
                    "classification_to",
                    "comment",
                ],
            )

        # Search the associated tree file
        tree_filename = conc_file.name.replace("conc_", "tree_").replace("_bonsai", "")
        tree_file = directory / tree_filename
        tree_file_bonsai = directory / "tree_bonsai.csv"

        # Check the columns and levels of the tree file
        df_tree = assert_columns(tree_file, ["code", "name"])

        df_tree_bonsai = assert_columns(tree_file_bonsai, ["code", "name"])

        # Check if source classification codes are in the associated tree file
        suffix = "_from" if "_bonsai.csv" in conc_file.name else "_to"

        if "activitytype" in directory.name:
            category = "activitytype"
        elif "flowobject" in directory.name:
            category = "flowobject"

        col = category + suffix
        missing_codes = set(df_conc[col]) - set(df_tree["code"])
        missing_codes = {code for code in missing_codes if pd.notna(code)}
        if missing_codes:
            logger.warning(
                f"Codes missing in '{category}/{tree_file.name}' (indicated by '{category}/{conc_file.name}'): {missing_codes}"
            )

        suffix = "_to" if "_bonsai.csv" in conc_file.name else "_from"
        col = category + suffix
        missing_codes = set(df_conc[col]) - set(df_tree_bonsai["code"])
        missing_codes = {code for code in missing_codes if pd.notna(code)}
        if missing_codes:
            logger.warning(
                f"Codes missing in '{category}/{tree_file_bonsai.name}' (indicated by '{category}/{conc_file.name}'): {missing_codes}"
            )


@pytest.mark.parametrize("directory", [flow_path], ids=["flow"])
def test_concpair_files(directory):
    # Check conc files
    conc_files = list(directory.glob("concpair_*_bonsai.csv")) + list(
        directory.glob("concpair_bonsai_*.csv")
    )
    for conc_file in conc_files:
        df_conc = assert_columns(
            conc_file,
            [
                "activitytype_from",
                "flowobject_from",
                "activitytype_to",
                "flowobject_to",
                "classification_from",
                "classification_to",
                "comment",
            ],
        )

        # Search the associated tree file
        tree_filename = conc_file.name.replace("concpair_", "tree_").replace(
            "_bonsai", ""
        )
        tree_file_act = directory / "activitytype" / tree_filename
        tree_file_act_bonsai = directory / "activitytype" / "tree_bonsai.csv"

        tree_file_flowobj = directory / "flowobject" / tree_filename
        tree_file_flobobj_bonsai = directory / "flowobject" / "tree_bonsai.csv"

        # Check the columns and levels of the tree file
        df_tree_act = assert_columns(tree_file_act, ["code", "name"])
        df_tree_flowobj = assert_columns(tree_file_flowobj, ["code", "name"])

        df_tree_flowobj_bonsai = assert_columns(
            tree_file_flobobj_bonsai, ["code", "name"]
        )
        df_tree_act_bonsai = assert_columns(tree_file_act_bonsai, ["code", "name"])

        # Check if source classification codes are in the associated tree file
        suffix = "_from" if "_bonsai.csv" in conc_file.name else "_to"
        col = "activitytype" + suffix
        missing_codes = set(df_conc[col]) - set(df_tree_act["code"])
        missing_codes = {code for code in missing_codes if pd.notna(code)}
        if missing_codes:
            logger.warning(
                f"Codes missing in activitytype/{tree_file_act.name}: {missing_codes}"
            )

        suffix = "_to" if "_bonsai.csv" in conc_file.name else "_from"
        col = "activitytype" + suffix
        missing_codes = set(df_conc[col]) - set(df_tree_act_bonsai["code"])
        missing_codes = {code for code in missing_codes if pd.notna(code)}
        if missing_codes:
            logger.warning(
                f"Codes missing in activitytype/{tree_file_act_bonsai.name}: {missing_codes}"
            )

        # Check if source classification codes are in the associated tree file
        suffix = "_from" if "_bonsai.csv" in conc_file.name else "_to"
        col = "flowobject" + suffix
        missing_codes = set(df_conc[col]) - set(df_tree_flowobj["code"])
        missing_codes = {code for code in missing_codes if pd.notna(code)}
        if missing_codes:
            logger.warning(
                f"Codes missing in flowobject/{tree_file_flowobj.name}: {missing_codes}"
            )

        # Check if source classification codes are in the associated tree file
        suffix = "_to" if "_bonsai.csv" in conc_file.name else "_from"
        col = "flowobject" + suffix
        missing_codes = set(df_conc[col]) - set(df_tree_flowobj_bonsai["code"])
        missing_codes = {code for code in missing_codes if pd.notna(code)}
        if missing_codes:
            logger.warning(
                f"Codes missing in flowobject/{tree_file_flobobj_bonsai.name}: {missing_codes}"
            )


def test_bonsai_products_to_markets_codes_exist():
    bonsai_products_to_markets_df = pd.read_csv(
        flow_path / "bonsai_sut_products_to_markets.csv"
    )

    activitytype_df = pd.read_csv(activitytype_path / "tree_bonsut.csv")
    flowobject_df = pd.read_csv(flowobject_path / "tree_bonsut.csv")
    # Extract unique values
    flowobject_values = set(bonsai_products_to_markets_df["flowobject"])
    activitytype_values = set(bonsai_products_to_markets_df["activitytype"])

    # Extract unique values from flowobject and activitytype CSVs (code + alias_code)
    flowobject_codes = set(flowobject_df["code"].dropna())
    activitytype_codes = set(activitytype_df["code"].dropna())

    # Assertions
    missing_flowobjects = flowobject_values - flowobject_codes
    missing_activitytypes = activitytype_values - activitytype_codes
    if missing_flowobjects:
        logger.warning(f"Missing flowobjects {missing_flowobjects}")
    if missing_activitytypes:
        logger.warning(f"Missing activitytypes {missing_activitytypes}")

    # assert not missing_flowobjects, f"Missing flowobject codes: {missing_flowobjects}"
    # assert (
    #    not missing_activitytypes
    # ), f"Missing activitytype codes: {missing_activitytypes}"


def test_bonsai_bonsut_correspondence_activitytype():
    df_B = pd.read_csv(activitytype_path / "tree_bonsai.csv")
    df_map = pd.read_csv(activitytype_path / "conc_bonsut_bonsai.csv")

    root_nodes_to_check = ["ai", "at"]

    # Step 1: Build parent and child mappings
    parent_lookup_B = dict(zip(df_B["code"], df_B["parent_code"]))
    children_lookup_B = df_B.groupby("parent_code")["code"].apply(list).to_dict()

    # Step 2: Traverse tree starting from specified roots to get all reachable nodes
    def dfs_collect_reachable_nodes(root, children_map):
        stack = [root]
        reachable = set()
        while stack:
            node = stack.pop()
            if node not in reachable:
                reachable.add(node)
                stack.extend(children_map.get(node, []))
        return reachable

    # Collect all nodes under specified roots
    reachable_nodes_B = set()
    for root in root_nodes_to_check:
        reachable_nodes_B |= dfs_collect_reachable_nodes(root, children_lookup_B)

    # Step 3: Identify leaf nodes among reachable nodes
    leaf_nodes_B = [code for code in reachable_nodes_B if code not in children_lookup_B]

    # Step 4: Get mapped B codes
    mapped_B_codes = set(df_map["activitytype_to"])

    # Step 5: Check if each reachable leaf is covered by a mapped ancestor
    def is_covered(code):
        while code is not None:
            if code in mapped_B_codes:
                return True
            code = parent_lookup_B.get(code)
        return False

    # Step 6: Find uncovered reachable leaf nodes
    uncovered_leaf_nodes = [code for code in leaf_nodes_B if not is_covered(code)]

    # Output
    if uncovered_leaf_nodes:
        logger.warning("Uncovered leaf nodes in tree_bonsai for activitytype:")
        logger.warning(uncovered_leaf_nodes)
    else:
        logger.info(
            "All leaf nodes in tree_bonsai for activitytype are covered by the correspondence."
        )


def test_bonsai_bonsut_correspondence_flowobject():
    df_B = pd.read_csv(flowobject_path / "tree_bonsai.csv")
    df_map = pd.read_csv(flowobject_path / "conc_bonsut_bonsai.csv")

    root_nodes_to_check = ["fi", "ft"]

    # Step 1: Build parent and child mappings
    parent_lookup_B = dict(zip(df_B["code"], df_B["parent_code"]))
    children_lookup_B = df_B.groupby("parent_code")["code"].apply(list).to_dict()

    # Step 2: Traverse tree starting from specified roots to get all reachable nodes
    def dfs_collect_reachable_nodes(root, children_map):
        stack = [root]
        reachable = set()
        while stack:
            node = stack.pop()
            if node not in reachable:
                reachable.add(node)
                stack.extend(children_map.get(node, []))
        return reachable

    # Collect all nodes under specified roots
    reachable_nodes_B = set()
    for root in root_nodes_to_check:
        reachable_nodes_B |= dfs_collect_reachable_nodes(root, children_lookup_B)

    # Step 3: Identify leaf nodes among reachable nodes
    leaf_nodes_B = [code for code in reachable_nodes_B if code not in children_lookup_B]

    # Step 4: Get mapped B codes
    mapped_B_codes = set(df_map["flowobject_to"])

    # Step 5: Check if each reachable leaf is covered by a mapped ancestor
    def is_covered(code):
        while code is not None:
            if code in mapped_B_codes:
                return True
            code = parent_lookup_B.get(code)
        return False

    # Step 6: Find uncovered reachable leaf nodes
    uncovered_leaf_nodes = [code for code in leaf_nodes_B if not is_covered(code)]

    # Output
    if uncovered_leaf_nodes:
        logger.warning("Uncovered leaf nodes in tree_bonsai for flowobject:")
        logger.warning(uncovered_leaf_nodes)
    else:
        logger.info(
            "All leaf nodes in tree_bonsai for flowobject are covered by the correspondence."
        )



import re
from importlib.resources import files

import pandas as pd
from dataio.resources import CSVResourceRepository

paths = [
    "flow/activitytype/",
    "location/",
    "dataquality/",
    "uncertainty/",
    "time/",
    "flow/flowobject/",
    "flow/",
]


source = files("classifications.data")


def test_dataio():
    for p in paths:
        path_to_resource = source.joinpath(p)
        repo = CSVResourceRepository(db_path=path_to_resource)
        resources = repo.list_available_resources()
        for r in resources:
            if r.name[0].isdigit():
                raise ValueError(
                    f"File name '{r.name}' starts with number in resource '{path_to_resource}'."
                )
            if not re.match(r"^[a-zA-Z0-9_]*$", r.name):
                raise ValueError(
                    f"File name '{r.name}' contains special character in resource '{path_to_resource}'."
                )
            df = repo.get_dataframe_for_task(r.name)
            if type(df) != pd.DataFrame:
                raise TypeError(
                    f"File '{r.name}' is not a pandas.DataFrame. Check location paths in the '{path_to_resource}/resource.csv'."
                )



import argparse
import csv
import os
import shutil
import tempfile
from collections import Counter, defaultdict
from logging import getLogger
from pathlib import Path

import classifications

ROOT_PATH = Path(os.path.dirname(__file__))

logger = getLogger("root")


def build_mapping_dict(rows, source_name, target_name):
    source_to_targets = defaultdict(set)
    target_to_sources = defaultdict(set)

    for row in rows:
        source = row[source_name]
        target = row[target_name]
        source_to_targets[source].add(target)
        target_to_sources[target].add(source)

    return source_to_targets, target_to_sources


# Define a function to determine the comment based on the counts
def get_comment(source, target, source_to_targets, target_to_sources):
    targets = source_to_targets.get(source, set())
    sources = target_to_sources.get(target, set())

    if len(targets) == 1 and len(sources) == 1:
        return "one-to-one correspondence"
    elif len(targets) > 1 and len(sources) == 1:
        # Check if the *other* targets also have multiple sources
        if any(len(target_to_sources[other_target]) > 1 for other_target in targets if other_target != target):
            return "ambiguous one-to-many correspondence"
        return "one-to-many correspondence"
    elif len(targets) == 1 and len(sources) > 1:
        # Check if the *source* targets also have multiple sources
        if any(len(source_to_targets[other_source]) > 1 for other_source in sources if other_source != source):
            return "ambiguous many-to-one correspondence" 
        return "many-to-one correspondence"
    elif len(targets) > 1 and len(sources) > 1:
        return "many-to-many correspondence"
    return ""




skos_uri_dict = {
    "one-to-one correspondence": "http://www.w3.org/2004/02/skos/core#exactMatch",
    "one-to-many correspondence": "http://www.w3.org/2004/02/skos/core#narrowMatch",
    "ambiguous one-to-many correspondence": "http://www.w3.org/2004/02/skos/core#narrowMatch",
    "many-to-one correspondence": "http://www.w3.org/2004/02/skos/core#broadMatch",
    "ambiguous many-to-one correspondence": "http://www.w3.org/2004/02/skos/core#broadMatch",
    "many-to-many correspondence": "http://www.w3.org/2004/02/skos/core#relatedMatch",
    "": "",
}


def add_mapping_comment(csv_file, overwrite="True"):
    """Helper to determine the mapping type.

    Attention!
    Only use overwrite='True' if only the most detailed codes of a classification are mapped.
    If parent codes are included in the correspondence table, the automated mapping would lead to wrong results.

    Arguments
    ---------
    csv_file: string
        path to csv file
    overwrite: string
        'True' to automate the mapping
        'False' if comment the column already includes the mapping type
    """
    if not os.path.exists(csv_file):
        logger.warning(f"File {csv_file} not found. Skipping...")
        return

    with open(csv_file, mode="r", newline="", encoding="utf-8") as file:
        reader = csv.DictReader(file)
        rows = list(reader)
        column_names = reader.fieldnames

    if not rows or not column_names:
        logger.warning(f"No valid data found in {csv_file}. Skipping...")
        return

    source_name, target_name = column_names[
        :2
    ]  # First two columns assumed as source & target

    # Count occurrences in advance for optimization
    source_to_targets, target_to_sources = build_mapping_dict(rows, source_name, target_name)

    if overwrite == "True":
        logger.warning(
            "Comment column, and thus the mapping type will be newly determined."
        )

        # Process rows
        for row in rows:
            source, target = row.get(source_name, ""), row.get(target_name, "")
            if source and target:
                row["comment"] = get_comment(
                    source, target, source_to_targets, target_to_sources
                )

                row["skos_uri"] = skos_uri_dict[row["comment"]]
            else:
                logger.warning(f"Skipping row with missing data: {row}")

    elif overwrite == "False":
        # Process rows
        for row in rows:
            source, target = row.get(source_name, ""), row.get(target_name, "")
            if source and target:
                row["skos_uri"] = skos_uri_dict[row["comment"]]
            else:
                logger.warning(f"Skipping row with missing data: {row}")

    # Write updated data to a temp file
    # Ensure "comment" and "skos_uri" are only added if they don't already exist
    if "comment" not in column_names:
        column_names.append("comment")
    if "skos_uri" not in column_names:
        column_names.append("skos_uri")

    fieldnames = column_names

    with tempfile.NamedTemporaryFile(
        mode="w", delete=False, newline="", encoding="utf-8"
    ) as temp_file:
        writer = csv.DictWriter(temp_file, fieldnames=fieldnames, lineterminator="\n")
        writer.writeheader()
        writer.writerows(rows)

    # Replace the original file
    shutil.move(temp_file.name, csv_file)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Process a specific CSV file to add mapping comments."
    )
    parser.add_argument("csv_file", help="Path to the CSV file you want to process.")
    parser.add_argument(
        "overwrite",
        help="Boolean, if True comment column will be overwritten, and thus the mapping type newly determined.",
    )

    args = parser.parse_args()

    if os.path.exists(args.csv_file):
        add_mapping_comment(args.csv_file, args.overwrite)
    else:
        logger.error(f"File {args.csv_file} not found.")



import os
import shutil
import tempfile
from pathlib import Path

import pandas as pd
import pytest

import classifications
from classifications import _mapping_type

test_file = Path(os.path.dirname(__file__)).parent / "tests/data/conc_test.csv"
test_tree_file = (
    Path(os.path.dirname(__file__)).parent / "tests/data/tree_test_correspondence.csv"
)


def test_mapping_type():
    tree_test = pd.read_csv(test_tree_file)

    classifications.flowobject.datapackage.tree_test_correspondence = tree_test

    _mapping_type.add_mapping_comment(test_file)
    df_result = pd.read_csv(test_file)
    assert df_result["comment"].to_list() == [
        "ambiguous one-to-many correspondence",
        "many-to-many correspondence",
        "one-to-one correspondence",
        "many-to-one correspondence",
        "many-to-one correspondence",
        "ambiguous one-to-many correspondence",
        "many-to-many correspondence",
        "many-to-many correspondence",
        "many-to-many correspondence",
    ]



import importlib.resources
import os
import re
import warnings
from collections import defaultdict
from functools import lru_cache
from logging import getLogger
from pathlib import Path

import pandas
from anytree import Node, RenderTree

from ._mapping_type import get_comment, skos_uri_dict

logger = getLogger("root")

ROOT_PATH = Path(os.path.dirname(__file__))

activitytype_path = "data/flow/activitytype/"
location_path = "data/location/"
dataquality_path = "data/dataquality/"
uncertainty_path = "data/uncertainty/"
time_path = "data/time/"
flowobject_path = "data/flow/flowobject/"
flow_path = "data/flow/"
unit_monetary_path = "data/unit/monetary/"
unit_physical_path = "data/unit/physical/"
unit_magnitude_path = "data/unit/magnitude/"

# Lookup function for pandas DataFrame
def lookup(self, keyword):
    """Filter the DataFrame based on the keyword in the "name" column"""
    filtered_df = self[self["name"].str.contains(keyword, case=False)]
    return filtered_df


def get_children(
    self,
    parent_codes,
    deep=True,
    return_parent=False,
):
    """
    Get descendants (direct and indirect) for a list of parent_codes.

    Parameters
    ----------
    parent_codes: str or list
        A single parent_code or a list of parent_codes for which descendants are to be fetched.
    deep: bool, optional
        If True, fetch all descendants recursively. If False, fetch only direct children. Default is True.
    return_parent: bool, optional
        If True, include the parent codes in the returned DataFrame. Default is False.

    Returns
    -------
    pandas.DataFrame
        A DataFrame containing rows with descendants of the specified parent_codes.
    """
    if not isinstance(self, pandas.DataFrame):
        raise TypeError("The object must be a pandas DataFrame.")

    if not {"code", "parent_code"}.issubset(self.columns):
        raise KeyError("DataFrame must contain 'code' and 'parent_code' columns.")

    if isinstance(parent_codes, str):
        parent_codes = [parent_codes]
    elif not isinstance(parent_codes, (list, set, tuple)):
        raise TypeError("parent_codes must be a string or a list-like object.")

    parent_codes = set(parent_codes)

    if deep:
        to_explore = set(parent_codes)
        all_descendants = set()

        while to_explore:
            current_children = self[self["parent_code"].isin(to_explore)]
            new_descendants = set(current_children["code"]) - all_descendants

            if not new_descendants:
                break  # Exit loop if no new children found

            all_descendants.update(new_descendants)
            to_explore = new_descendants  # Continue exploring new children

        if not all_descendants:
            logger.debug(f"No children found for {parent_codes}")

        df = self[self["code"].isin(all_descendants)]
    else:
        df = self[self["parent_code"].isin(parent_codes)]
        if df.empty:
            df = self[self["code"].isin(parent_codes)]

    # Include parent codes if requested
    if return_parent:
        df = pandas.concat(
            [self[self["code"].isin(parent_codes)], df]
        ).drop_duplicates()

    return CustomDataFrame(df)


def create_conc(df_A, df_B, source="", target=""):
    """Create new concordance based on two other tables.

    Argument
    --------
    df_A : pandas.DataFrame
        concordance table A
        with mapping from "x" to "y"
    df_B : pandas.DataFrame
        concordance table B
        with mapping from "y" to "z"
    target : str
        classification name that specifies "x"
    source : str
        classification name that specifies "z"

    Returns
    -------
    pandas.DataFrame
        concordance table with mapping form "x" to "z"
    """
    if "activitytype_to" in df_B.columns and "flowobjet_to" in df_B.columns:
        raise NotImplementedError("Concpair tables not allowed")
    elif "activitytype_to" in df_A.columns and "activitytype_to" in df_B.columns:
        column_prefix = "activitytype"
    elif "flowobject_to" in df_A.columns and "flowobject_to" in df_B.columns:
        column_prefix = "flowobject"

    merged = pandas.merge(df_A, df_B, on=f"{column_prefix}_to", suffixes=("_A", "_B"))

    # Create the resulting DataFrame with required columns
    result = pandas.DataFrame(
        {
            f"{column_prefix}_from": merged[f"{column_prefix}_from_A"],
            f"{column_prefix}_to": merged[f"{column_prefix}_from_B"],
            "classification_from": source,  # Fixed value from A
            "classification_to": target,  # Fixed value for result
        }
    )

    # Drop duplicate pairs of source and target
    new_mapping = result.drop_duplicates(
        subset=[
            f"{column_prefix}_from",
            f"{column_prefix}_to",
            "classification_from",
            "classification_to",
        ]
    )

    # Calculate the counts of each source and target in the merged DataFrame
    source_counts = new_mapping[f"{column_prefix}_from"].value_counts().to_dict()
    target_counts = new_mapping[f"{column_prefix}_to"].value_counts().to_dict()
    # Apply the get_comment function to each row
    # Build relationship dictionaries first
    source_to_targets = defaultdict(set)
    target_to_sources = defaultdict(set)

    for _, row in new_mapping.iterrows():
        source = row[f"{column_prefix}_from"]
        target = row[f"{column_prefix}_to"]
        if source and target:
            source_to_targets[source].add(target)
            target_to_sources[target].add(source)

    # Apply revised comment logic using get_true_comment
    new_mapping["comment"] = new_mapping.apply(
        lambda row: get_comment(
            row[f"{column_prefix}_from"],
            row[f"{column_prefix}_to"],
            source_to_targets,
            target_to_sources,
        ),
        axis=1,
    )

    new_mapping["skos_uri"] = new_mapping["comment"].map(skos_uri_dict)

    new_mapping = new_mapping[
        [
            f"{column_prefix}_from",
            f"{column_prefix}_to",
            "classification_from",
            "classification_to",
            "comment",
            "skos_uri",
        ]
    ]
    new_mapping = new_mapping.reset_index(drop=True)
    return new_mapping


def _get_concordance_file(file_path):
    try:
        # Read the concordance CSV into a DataFrame
        return pandas.read_csv(file_path, dtype=str)
        # return multiple_dfs
    except FileNotFoundError:
        return None
    except Exception as e:
        logger.error(f"Error reading concordance file: {e}")
        return None


def get_concordance(from_classification, to_classification, category):
    """
    Get the concordance DataFrame based on the specified classifications.
    Parameters
    ----------
    from_classification: str
        The source classification name (e.g., "bonsai").
    to_classification: str
        The target classification name (e.g., "nace_rev2").
    category: str
        category to look in (e.g. location, activitytype)
    Returns
    -------
    pd.DataFrame
        The concordance DataFrame if 1 file is found; otherwise, a dict of DataFrames.
    """
    # Construct the file name
    fitting_file_names = [
        f"conc_{from_classification}_{to_classification}.csv",
        f"concpair_{from_classification}_{to_classification}.csv",
    ]
    reversed_file_names = [
        f"conc_{to_classification}_{from_classification}.csv",
        f"concpair_{to_classification}_{from_classification}.csv",
    ]
    path_dict = {
        "activitytype": "data/flow/activitytype/",
        "location": "data/location/",
        "dataquality": "data/dataquality/",
        "uncertainty": "data/uncertainty/",
        "time": "data/time/",
        "flowobject": "data/flow/flowobject/",
        "flow": "data/flow/",
        "currency": "data/currency/",
    }

    file_path = path_dict[f"{category}"]
    file_paths = [file_path]

    multiple_dfs = {}
    for f in file_paths:
        for n in fitting_file_names:
            file_path = ROOT_PATH.joinpath(f, n)
            df = _get_concordance_file(file_path)
            if not df is None:
                multiple_dfs[f"{f}"] = df
        for n in reversed_file_names:
            file_path = ROOT_PATH.joinpath(f, n)
            df = _get_concordance_file(file_path)
            if not df is None:
                # Renaming columns
                new_columns = {}
                for col in df.columns:
                    if "_from" in col:
                        new_columns[col] = col.replace("_from", "_to")
                    elif "_to" in col:
                        new_columns[col] = col.replace("_to", "_from")
                df.rename(columns=new_columns, inplace=True)

                # Changing the comment column
                df["comment"] = df["comment"].replace(
                    {
                        "one-to-many correspondence": "many-to-one correspondence",
                        "many-to-one correspondence": "one-to-many correspondence",
                    }
                )
                df["skos_uri"] = df["skos_uri"].replace(
                    {
                        "http://www.w3.org/2004/02/skos/core#narrowMatch": "http://www.w3.org/2004/02/skos/core#broadMatch",
                        "http://www.w3.org/2004/02/skos/core#broadMatch": "http://www.w3.org/2004/02/skos/core#narrowMatch",
                    }
                )
                multiple_dfs[f"{f}"] = df

    if len(multiple_dfs) == 1:
        return multiple_dfs[next(iter(multiple_dfs))]
    elif len(multiple_dfs) > 1:
        return multiple_dfs
    else:
        raise FileNotFoundError(
            f"No concordance for '{from_classification}' and '{to_classification}' found."
        )


def get_tree(for_classification, ctype):
    """
    Get the tree table as DataFrame for a classification
    Parameters
    ----------
    for_classification: str
        Name of the requested classification (e.g., "nace_rev2").
    Returns
    -------
    pd.DataFrame
        The tree as DataFrame
    """
    # Search all file_paths
    if ctype in ["activitytype", "flowobject"]:
        ctype = f"flow/{ctype}"
    file_path = ROOT_PATH.joinpath("data", ctype, f"tree_{for_classification}.csv")
    return _get_concordance_file(file_path)


def print_tree(self, toplevelcode):
    """Print the tree structure for a given code.

    Bold text represent sub-categories which are included when applying it in the Bonsai SUT.
    Italic text represent sub-categories which are not included, since these are separate codes in the Bonsai SUT.

    """
    all_codes = self.get_children(
        toplevelcode,
        deep=True,
        return_parent=True,
    )
    sut_codes = self.get_children(
        toplevelcode,
        deep=True,
        return_parent=True,
    )
    # Create nodes from the data
    nodes = {}
    for _, row in all_codes.iterrows():
        nodes[row["code"]] = Node(
            row["code"], parent=nodes.get(row["parent_code"]), descript=row["name"]
        )

    italic_codes = set(sut_codes["code"])  # Set of codes to make italic
    for pre, fill, node in RenderTree(nodes[toplevelcode]):
        if node.name in italic_codes:
            print(f"{pre}\033[1m{node.name} - {node.descript}\033[0m")  # Italicize text
        else:
            print(f"{pre}\033[3m{node.name} - {node.descript}\033[0m")


def convert_name(self, name):
    """
    Convert a name or list of names based on regex matching rules defined in the DataFrame.

    This method searches for matches using the 'location_from' regex patterns
    and returns the corresponding 'location_to' values for each match.

    Parameters:
    ----------
    name : str or list of str
        A single name (string) or a list of names to convert.

    Returns:
    -------
    list or list of lists or None
        - If a single name is provided, returns a list of matched 'location_to' values,
          or None if no matches are found.
        - If a list of names is provided, returns a list where each element is either:
            * a list of matched 'location_to' values for the corresponding input name, or
            * None if no matches are found for that name.

    Raises:
    ------
    NotImplementedError
        If the 'classification_from' column does not contain the value 'regex'.

    Example:
    -------
    >>> conc_bonsai_regex.convert_name("USA")
    'US'

    >>> conc_bonsai_regex.convert_name(["Great Britain", "China"])
    ['GB', 'CN']
    """
    if "regex" not in self["classification_from"].values:
        raise NotImplementedError("Method not applicable, since no 'regex' column.")

    def get_matches(single_name):
        matches = []
        for _, row in self.iterrows():
            if re.search(row["location_from"], single_name):
                matches.append(row["location_to"])
        if not matches:
            return None
        if len(matches) == 1:
            return matches[0]
        return matches

    if isinstance(name, list):
        return [get_matches(n) for n in name]
    else:
        return get_matches(name)


def nearest_other_code(self, code, correspondence_to_other):
    """
    Return the nearest OTHER code that corresponds to the given code of a classification tree.

    Searches for any descendant (child, grandchild, etc.) with the same name
    and in_final_sut == True. If not found, moves up to ancestors.

    :param self: Pandas DataFrame containing the hierarchical tree (with 'code', 'parent_code', 'name', 'in_final_sut')
    :param code: The code to check
    :param correspondence_to_other: Pandas DataFrame with corresponding mapping to another classification
    :return tuple: (original_code, corresponding_sut_code) or None if no match is found
    """
    visited_codes = set()

    correspondence_to_other

    def get_sut_mapping(original_code):
        code_str = str(original_code)

        df = correspondence_to_other.copy()

        # List of possible from/to column pairs
        column_pairs = [
            ("flowobject_from", "flowobject_to"),
            ("activitytype_from", "activitytype_to"),
            ("location_from", "location_to"),
        ]

        for col_from, col_to in column_pairs:
            # Normalize columns
            if col_from in df.columns and col_to in df.columns:
                df[col_from] = df[col_from].astype(str)
                df[col_to] = df[col_to].astype(str)

                # Forward match: code == to → return from
                forward = df[df[col_to] == code_str]
                if not forward.empty:
                    return forward[col_from].values[0]

                # Reverse match: code == from → return to
                reverse = df[df[col_from] == code_str]
                if not reverse.empty:
                    return reverse[col_to].values[0]

        return None

    def find_valid_descendant(current_code, name):
        descendants = self[self["parent_code"] == current_code]
        matches = []
        for _, descendant in descendants.iterrows():
            sut_code = get_sut_mapping(descendant["code"])
            if sut_code:
                matches.append((descendant["code"], sut_code))
            child_matches = find_valid_descendant(descendant["code"], name)
            matches.extend(child_matches)
        return matches

    while True:
        if code in visited_codes:
            return None  # Avoid cycles
        visited_codes.add(code)

        row = self[self["code"] == code]
        if row.empty:
            return None

        name = row["name"].values[0]

        # Check descendants
        descendant_result = find_valid_descendant(code, name)
        if descendant_result:
            return descendant_result

        # Check current code
        if row["code"].values[0] == code:
            sut_code = get_sut_mapping(code)
        if sut_code:
            return code, sut_code

        # Move to parent
        parent_code = row["parent_code"].values[0]
        if pandas.isna(parent_code) or parent_code == "":
            return None

        code = parent_code


# Subclass pandas DataFrame
class CustomDataFrame(pandas.DataFrame):
    lookup = lookup
    get_children = get_children
    print_tree = print_tree
    convert_name = convert_name
    nearest_other_code = nearest_other_code


@lru_cache(maxsize=1)
def get_currency_rate_df():
    with importlib.resources.files("classifications.data.unit.monetary").joinpath(
        "fact_currency_per_usd.csv"
    ).open("r") as file:
        return pandas.read_csv(file)


@lru_cache(maxsize=1)
def get_deflator_df():
    with importlib.resources.files("classifications.data.unit.monetary").joinpath(
        "fact_deflator_deu.csv"
    ).open("r") as file:
        return pandas.read_csv(file)


def convert_currency(value_from, from_currency, to_currency, ureg):
    df_curr_rate = get_currency_rate_df()
    df_defl_factor = get_deflator_df()

    # split up the currency
    from_currency, from_year = from_currency.split("_", 1)
    to_currency, to_year = to_currency.split("_", 1)
    from_year = int(from_year)
    to_year = int(to_year)

    row_from = df_curr_rate[
        (df_curr_rate["unit"] == f"{from_currency}/USD")
        & (df_curr_rate["time"] == from_year)
    ]
    if row_from.empty:
        raise ValueError(f"No rate for {from_currency}/USD in from_year {from_year}")
    rate_from_per_usd = row_from["value"].iloc[0]

    # Get rate of to_currency per USD
    row_to = df_curr_rate[
        (df_curr_rate["unit"] == f"{to_currency}/USD")
        & (df_curr_rate["time"] == from_year)
    ]
    if row_to.empty:
        raise ValueError(f"No rate for {to_currency}/USD in to_year {to_year}")
    rate_to_per_usd = row_to["value"].iloc[0]

    # Convert: from/USD divided by to/USD
    rate_from_per_to = rate_from_per_usd / rate_to_per_usd

    # Adjust with the deflator
    deflator_to = df_defl_factor[df_defl_factor["time"] == to_year]["value"].iloc[0]
    deflator_from = df_defl_factor[df_defl_factor["time"] == from_year]["value"].iloc[0]

    new_currency = rate_from_per_to * deflator_to / deflator_from

    return value_from / new_currency * (getattr(ureg, f"{to_currency}_{to_year}"))



# -*- coding: utf-8 -*-
"""
Created on Fri Mar  7 08:20:13 2025

@author: nielenssvan
"""
import pandas as pd
import classifications

def find_missing_alias(tree: pd.DataFrame, code: str, check_children=False):
    """Find codes in a Bonsai classification tree with aliases.
    Print the minimal set of codes that need an alias to make
    the coverage of aliases complete.
    """
    if not check_children and tree.loc[tree.code == code, "alias_code"].notna().all():
        return True
    elif not any(tree.parent_code == code):
        return False
    children = tree.loc[tree.parent_code == code, "code"]
    found = {True: [], False: []}
    for child in children:
        answer = find_missing_alias(tree, child)
        found[answer].append(child)
    if tree.loc[tree.code == code, "alias_code"].notna().all():
        return True
    if not any(found[False]):
        return True
    elif any(found[True]):
        print(f"Add alias to: {found[False]}")
        return True
    else:
        return False
        
def find_missing_bonsut(field: str, check_children=False):
    """Combine the Bonsai tree and the Bonsut codes.
    Then check if the coverage of Bonsut codes is complete.
    """
    if field == "product":
        tree = classifications.flowobject.datapackage.tree_bonsai
        bonsut_corr = classifications.flowobject.datapackage.conc_bonsut_bonsai
        bonsut_corr = bonsut_corr[["flowobject_from", "flowobject_to"]]
        parent_codes = ["fi"]
    elif field == "activity":
        tree = classifications.activitytype.datapackage.tree_bonsai
        bonsut_corr = classifications.activitytype.datapackage.conc_bonsut_bonsai
        bonsut_corr = bonsut_corr[["activitytype_from", "activitytype_to"]]
        parent_codes = ["ai", "at"]
    else:
        print("Field not recognized. Expecting 'product' or 'activity'.")

    bonsut_corr.columns = ["alias_code", "code"]
    tree = tree[["code", "parent_code", "level"]].merge(bonsut_corr, "left")
    for code in parent_codes:
        result = find_missing_alias(tree, code, check_children)
        if result:
            print("No missing Bonsut codes identified.")
        if not check_children:
            print("Descendants only considered if parent has no Bonsut code.")



# -*- coding: utf-8 -*-
"""
Created on Mon Dec 30 13:29:20 2024
Utility functions for creating correspondence tables in the classification package

Author: Sander van Nielen
"""
from logging import getLogger
import os
from pathlib import Path

import classifications as classif
import pandas as pd

logger = getLogger("root")

#NOTE: these variables already exist in _utils.py
ROOT_PATH = Path(os.path.dirname(__file__))
activitytype_path = "data/flow/activitytype/"
flowobject_path = "data/flow/flowobject/"

def extrapolate_corr(external_cls: str, how="up|down", save: bool=False):
    """
    Extrapolate a correspondence table to higher/lower Bonsai classification levels.

    Parameters:
        external_cls (str): External classification, for which a
            correspondence exists to Bonsai codes
        how (str): Whether to extrapolate up or down the tree
        save (bool): Whether or not to save the result (overwrites original)

    Returns:
        corr (pd.DataFrame): Table with extended correspondences 
    """
    assert how in ["up", "down"], f"Unrecognized parameter value how={how}"
    # Get correspondence table
    corr = classif.get_concordance(external_cls, "bonsai")
    if "flowobject_from" in corr.columns:
        data_path = ROOT_PATH / flowobject_path
        from_col, to_col = "flowobject_from", "flowobject_to"
    elif "activitytype_from" in corr.columns:
        data_path = ROOT_PATH / activitytype_path
        from_col, to_col = "activitytype_from", "activitytype_to"
    # Find all bonsai codes (in to_col) that have an associated external code
    corr = corr[corr[from_col].notna() & corr[to_col].notna()]
    mapped = set(corr[to_col])
    # Merge the concordance table with the bonsai tree
    tree_columns = ["code", "name", "parent_code", "level"]
    tree = pd.read_csv(data_path / "tree_bonsai.csv")[tree_columns]
    
    for bonsai_code in mapped:
        if bonsai_code not in list(tree.code):
            logger.warning(f"{bonsai_code} is missing in {data_path}/tree_bonsai.csv")
            continue
        
        # Select the correspondences of bonsai_code
        new_ext = list(corr[corr[to_col]==bonsai_code][from_col])
        new_bonsai = []
        
        # Collect all parents/children of bonsai_code that are not yet mapped
        if how == "up":
            this_code = tree[tree.code==bonsai_code].parent_code.iloc[0]
            while (tree[tree.code==this_code].level.max() > 0) & (this_code not in mapped):
                new_bonsai += len(new_ext) * [this_code]
                this_code = tree[tree.code==this_code].parent_code.iloc[0]
        else:
            this_code = bonsai_code
            children = classif._utils.get_children(tree, this_code, deep=False)
            while len(children) > 0:
                missing = set(children.code) - set(corr[from_col])
                if len(children.unique()) > len(missing):
                    logger.warning(f"Some children of {bonsai_code} are defined unexpectedly.")
                elif any(missing):
                    new_bonsai += len(new_ext) * list(missing)
                children = classif._utils.get_children(tree, missing, deep=False)
        # Apply the external codes to the collected bonsai codes
        if any(new_bonsai):
            new_corr = pd.DataFrame({from_col: int(len(new_bonsai)/len(new_ext)) * new_ext,
                                     to_col: new_bonsai})
            corr = pd.concat([corr, new_corr])
    
    # Fill the classifications columns
    for col in ["classification_from", "classification_to"]:
        corr[col] = corr[col].iloc[0]
    if save:
        file_path = data_path / f"conc_{external_cls}_bonsai.csv"
        corr.to_csv(file_path, index=False)
    return corr

def extrapolate_down(external_cls: str, save: bool=False):
    """
    Extrapolate a correspondence table to higher Bonsai classification levels.

    Parameters:
        external_cls (str): External classification, for which a
            correspondence exists to Bonsai codes
        save (bool): Whether or not to save the result (overwrites original)

    Returns:
        corr (pd.DataFrame): Table with extended correspondences 
    """
    # Get correspondence table
    corr = classif.get_concordance(external_cls, "bonsai")
    if "flowobject_from" in corr.columns:
        data_path = ROOT_PATH / flowobject_path
        from_col, to_col = "flowobject_from", "flowobject_to"
    elif "activitytype_from" in corr.columns:
        data_path = ROOT_PATH / activitytype_path
        from_col, to_col = "activitytype_from", "activitytype_to"
    # Find all bonsai codes (in to_col) that have an associated external code
    corr = corr[corr[from_col].notna() & corr[to_col].notna()]
    mapped = corr[to_col].unique()
    # Merge the concordance table with the bonsai tree
    tree_columns = ["code", "name", "parent_code", "level"]
    tree = pd.read_csv(data_path / "tree_bonsai.csv")[tree_columns]
    
    for bonsai_code in mapped:
        # Select the correspondences of bonsai_code
        new_ext = list(corr[corr[to_col]==bonsai_code][from_col])
        new_bonsai = []
        if len(new_ext) == 0:
            logger.warning(f"{bonsai_code} is missing in {data_path}/tree_bonsai.csv")
            continue
        # Collect all children of bonsai_code that are not yet mapped
        this_code = bonsai_code
        children = classif._utils.get_children(tree, this_code, deep=False)
        while len(children) > 0:
            missing = set(children.code) - set(corr[from_col])
            if len(children.unique()) > len(missing):
                logger.warning(f"Some children of {bonsai_code} are defined unexpectedly.")
            elif any(missing):
                new_bonsai += len(new_ext) * list(missing)
            children = classif._utils.get_children(tree, missing, deep=False)
        # Apply the codes of the child to 
        if any(new_bonsai):
            new_corr = pd.DataFrame({from_col: int(len(new_bonsai)/len(new_ext)) * new_ext,
                                     to_col: new_bonsai})
            corr = pd.concat([corr, new_corr])
    
    # Fill the classifications columns
    for col in ["classification_from", "classification_to"]:
        corr[col] = corr.loc[0, col]
    if save:
        file_path = data_path / f"conc_{external_cls}_bonsai.csv"
        corr.to_csv(file_path, index=False)
    return corr

def extrapolate_bea():
    # Products
    corr = classif.flowobject.datapackage.conc_bea_prod_bonsai
    tree = classif.flowobject.datapackage.tree_bea_prod
    tree = tree[["code", "name", "parent_code", "level"]]
    # combi = tree.rename(columns={"code": "flowobject_from"}).merge(corr, "outer")
    # missing = set(tree.code) - set(corr.flowobject_from)
    
    # Find all bea codes (in from_col) that have an associated bonsai code
    from_col, to_col = "flowobject_from", "flowobject_to"
    corr = corr[corr[from_col].notna() & corr[to_col].notna()]
    mapped = set(corr[from_col])
    
    for bea_code in mapped:
        if bea_code not in list(tree.code):
            logger.warning(f"{bea_code} is missing in tree_bea_prod.csv")
            continue
        
        # Select the correspondences of bea_code
        new_bonsai = list(corr[corr[from_col]==bea_code][to_col])
        new_bea = []
        
        # Collect all parents/children of bea_code that are not yet mapped
        this_code = tree[tree.code==bea_code].parent_code.iloc[0]
        while (tree[tree.code==this_code].level.max() > 0) & (this_code not in mapped):
            new_bea += len(new_bonsai) * [this_code]
            this_code = tree[tree.code==this_code].parent_code.iloc[0]
        # Apply the bonsai codes to the collected bea codes
        if any(new_bea):
            new_corr = pd.DataFrame({to_col: int(len(new_bea)/len(new_bonsai)) * new_bonsai,
                                     from_col: new_bea})
            corr = pd.concat([corr, new_corr])
    
    # Fill the classifications columns
    for col in ["classification_from", "classification_to"]:
        corr[col] = corr[col].iloc[0]
    
    # Activities
    corr = classif.activitytype.datapackage.conc_bea_activ_bonsai
    tree = classif.activitytype.datapackage.tree_bea_activ
    combi = tree.rename(columns={"code": "activitytype_from"}).merge(corr, "outer")
    print(len(combi[combi.activitytype_to.isna()]))
    for code in combi[combi.activitytype_to.isna()].activitytype_from:
        child_rows = combi[combi.parent_code == code]
        if len(child_rows)==1 and child_rows.activitytype_to.any():
            combi.loc[combi.activitytype_from==code, "activitytype_to"] = child_rows.activitytype_to.values[0]
    print(len(combi[combi.activitytype_to.isna()]))
    for code in combi[combi.activitytype_to.isna()].activitytype_from:
        child_rows = combi[combi.parent_code == code]
        if len(child_rows)==1 and child_rows.activitytype_to.any():
            combi.loc[combi.activitytype_from==code, "activitytype_to"] = child_rows.activitytype_to.values[0]
        elif len(child_rows)>1 and child_rows.activitytype_to.any():
            match = list(child_rows.activitytype_to.dropna().unique())
            match.sort()
            match = ", ".join(match)
            combi.loc[combi.activitytype_from==code, "activitytype_to"] = match
    print(len(combi[combi.activitytype_to.isna()]))



import argparse
import os

import classifications


def main():
    parser = argparse.ArgumentParser(description="Disaggregate Bonsai Codes")
    parser.add_argument(
        "category", type=str, help="Bonsai category (e.g. activitytype or flowobject)"
    )
    parser.add_argument(
        "yaml_file", type=str, help="Path to the YAML file with disaggregation data"
    )
    parser.add_argument(
        "--output_dir",
        type=str,
        default="output",
        help="Directory to save the updated DataFrames",
    )

    args = parser.parse_args()

    if args.category == "activitytype":
        processor = classifications.activitytype.datapackage
    elif args.category == "flowobject":
        processor = classifications.flowobject.datapackage
    elif args.category == "location":
        processor = classifications.location.datapackage
    elif args.category == "uncertainty":
        processor = classifications.uncertainty.datapackage
    elif args.category == "dataquality":
        processor = classifications.dataquality.datapackage
    elif args.category == "time":
        processor = classifications.time.datapackage
    else:
        raise NotImplementedError(
            f"your choice '{args.category}' is not a Bonsai category"
        )

    # Call the disaggregate_bonsai method with the path to the YAML file
    result = processor.disaggregate_bonsai(args.yaml_file)

    # Create the output directory if it doesn't exist
    os.makedirs(args.output_dir, exist_ok=True)

    # Save each DataFrame in the result as a CSV file only if it has changed
    for attr_name in result:
        original_df = getattr(processor, attr_name)
        updated_df = result[attr_name]

        # Check if the DataFrame has changed
        if not updated_df.equals(original_df):
            output_path = os.path.join(args.output_dir, f"{attr_name}_updated.csv")
            updated_df.to_csv(output_path, index=False)
            print(f"Saved updated DataFrame '{attr_name}' to {output_path}")
        else:
            print(f"No changes detected for DataFrame '{attr_name}'. Skipping save.")


if __name__ == "__main__":
    main()



import os
from logging import getLogger
from pathlib import Path

import pandas as pd
import yaml
from pandas._libs.parsers import STR_NA_VALUES

from ._utils import (
    CustomDataFrame,
    activitytype_path,
    dataquality_path,
    flow_path,
    flowobject_path,
    location_path,
    time_path,
    uncertainty_path,
    unit_magnitude_path,
    unit_monetary_path,
    unit_physical_path,
)

logger = getLogger("root")

ROOT_PATH = Path(os.path.dirname(__file__))

accepted_na_values = STR_NA_VALUES - {"NA"}


class DataPackage:
    def to_dict(self):
        result = {}
        for attr_name in dir(self):
            attr_value = getattr(self, attr_name)
            if isinstance(attr_value, pd.DataFrame):
                result[attr_name] = attr_value
        return result

    def disaggregate_bonsai(self, codes_or_yaml):
        """Disaggregates the code in the bonsai tables.

        Currently only for activitytype and flowobject.

        Argument
        --------
        codes_or_yaml: dict or str
            A dictionary containing the disaggregation mappings or a string path to a YAML file.
            Example::

                disaggregations:
                  - old_code: "ai"
                    new_codes:
                      - code: "ai_0"
                        description: "foo"
                        mappings: {"nace_rev1": ["0.11", "0.12"]}
                      - code: "ai_1"
                        description: ""
                        mappings: {}

        Returns
        -------
        dict
            dict with updated bonsai tables
        """

        # Determine if the input is a dictionary or a YAML file path
        if isinstance(codes_or_yaml, dict):
            disaggregations = codes_or_yaml.get("disaggregations", [])
        elif isinstance(codes_or_yaml, str):
            # Load the YAML file
            with open(codes_or_yaml, "r") as file:
                yaml_content = yaml.safe_load(file)
                disaggregations = yaml_content.get("disaggregations", [])
        else:
            raise ValueError("Input must be a dictionary or a path to a YAML file")

        # Validate the structure of the YAML content
        if not isinstance(disaggregations, list):
            raise ValueError("Disaggregations should be a list of mappings")
        if not "old_code" and "new_codes" in disaggregations[0].keys():
            raise ValueError("Make sure that you provide 'old_code' and 'new_codes'.")

        result = {}
        for attr_name in dir(self):
            attr_value = getattr(self, attr_name)
            if isinstance(attr_value, pd.DataFrame):
                result[attr_name] = attr_value

        for attr_name in result:
            attr_value = getattr(self, attr_name)
            if isinstance(attr_value, pd.DataFrame) and attr_name.startswith(
                "tree_bonsai"
            ):
                # For each disaggregation in the YAML content
                for disaggregation in disaggregations:
                    old_code = disaggregation["old_code"]
                    new_codes = disaggregation["new_codes"]
                    if len(new_codes) == 1:
                        raise ValueError(
                            f"Only 1 new code provided. To disaggregate {old_code}, provide at least 2 new codes!"
                        )

                    # Replace the old code with each new code in a new row
                    new_rows = attr_value[attr_value["code"] == old_code].copy()

                    for new_code in new_codes:
                        new_rows["code"] = new_code["code"]
                        new_rows["name"] = new_code["description"]
                        new_rows["parent_code"] = old_code
                        new_rows["level"] = (
                            int(attr_value[attr_value["code"] == old_code]["level"]) + 1
                        )
                        new_rows["comment"] = ""

                        # Append the new rows to the DataFrame
                        attr_value = pd.concat(
                            [attr_value, new_rows], ignore_index=True
                        )

                    # Update the DataFrame in the instance with the modified one
                    result[attr_name] = attr_value

            if (
                isinstance(attr_value, pd.DataFrame)
                and attr_name.startswith("conc_")
                and "bonsai" in attr_name
            ):
                # For each disaggregation in the YAML content
                for disaggregation in disaggregations:
                    old_code = disaggregation["old_code"]
                    new_codes = disaggregation["new_codes"]

                    # Replace the old code with each new code in a new row
                    for new_code in new_codes:
                        other_classifications = new_code.get("mappings", {})

                        df2 = pd.DataFrame(data=None, columns=attr_value.columns)

                        if other_classifications:
                            for schema, codes_list in other_classifications.items():
                                if schema in attr_name:
                                    if "activitytype_to" in df2.columns:
                                        df2["activitytype_to"] = codes_list
                                    if "flowobject_to" in df2.columns:
                                        df2["flowobject_to"] = codes_list

                            if "activitytype_from" in df2.columns:
                                df2["activitytype_from"] = new_code["code"]
                            if "flowobject_from" in df2.columns:
                                df2["flowobject_from"] = new_code["code"]
                            df2["comment"] = ""
                            df2["skos_uri"] = ""

                            # Append the new rows to the DataFrame
                            attr_value = pd.concat([attr_value, df2], ignore_index=True)

                            # Remove old parent code from concordance DataFrame (only mapping of most detailed category required)
                            if "activitytype_from" in attr_value.columns:
                                attr_value.drop(
                                    attr_value[
                                        # attr_value["tree_bonsai"] == f"{old_code}"
                                        attr_value["activitytype_from"]
                                        == f"{old_code}"
                                    ].index,
                                    inplace=True,
                                )
                            if "flowobject_from" in attr_value.columns:
                                attr_value.drop(
                                    attr_value[
                                        # attr_value["tree_bonsai"] == f"{old_code}"
                                        attr_value["flowobject_from"]
                                        == f"{old_code}"
                                    ].index,
                                    inplace=True,
                                )
                            attr_value.reset_index(drop=True, inplace=True)

                    # Update the DataFrame in the instance with the modified one
                    result[attr_name] = attr_value

        return result


class BonsaiTreeMaster:
    def __init__(self, code, parent_code, name, level, alias_code):
        self.code = code
        self.parent_code = parent_code
        self.name = name
        self.level = level
        self.alias_code = alias_code


class TreeMasterFlowobject(BonsaiTreeMaster):
    def __init__(
        self,
        code,
        parent_code,
        name,
        level,
        compartment,
        chemical_compound,
        default_unit,
        alias_code,
    ):
        super().__init__(code, parent_code, name, level, alias_code)
        self.compartment = compartment
        self.chemical_compound = chemical_compound
        self.default_unit = default_unit


class BonsaiDimMaster:
    def __init__(self, code, name, description, comment):
        self.code = code
        self.name = name
        self.description = description
        self.comment = comment


def list_csv_files_excluding(directory_path, exclude_file):
    # List all CSV files in the directory, excluding the specified file
    csv_files = [
        file for file in directory_path.glob("*.csv") if file.name != exclude_file
    ]
    return csv_files


def read_csv_files_as_dataframes(directory_path, exclude_file, prefix_dtype_mapping):
    csv_files = list_csv_files_excluding(directory_path, exclude_file)
    dataframes = {}
    for csv_file in csv_files:
        try:
            # Get the dtype dictionary based on the column name prefixes
            dtype_dict = get_dtype_dict_by_prefixes(csv_file, prefix_dtype_mapping)
            # Read each CSV file into a DataFrame with the specified dtypes
            file_stem = csv_file.stem  # Get the filename without the extension
            dataframes[file_stem] = pd.read_csv(
                csv_file, dtype=dtype_dict, na_values=[], keep_default_na=False
            )
        except Exception as e:
            logger.info(f"Error reading {csv_file.name}: {e}")
    return dataframes


def get_dtype_dict_by_prefixes(file_path, prefix_dtype_mapping):
    # Read the first row to get column names
    temp_df = pd.read_csv(file_path, nrows=0)
    column_names = temp_df.columns.tolist()

    # Create a dictionary mapping column names to dtypes based on prefix_dtype_mapping
    dtype_dict = {}
    for prefix, dtype in prefix_dtype_mapping.items():
        dtype_dict.update(
            {col: dtype for col in column_names if col.startswith(prefix)}
        )
    return dtype_dict


dtype_dict = {
    "code": "str",
    "parent_code": "str",
    "tree_": "str",
    "dim_": "str",
    "flowobject_from": "str",
    "flowobject_to": "str",
    "activitytype_from": "str",
    "activitytype_to": "str",
}


class ObjectNames:
    def __init__(self, code):
        self.code = code


class EmptyClass:
    def __init__(self) -> None:
        pass


class Resources:
    def __init__(self, path_to_resource):
        self.datapackage = DataPackage()
        self.bonsai = DataPackage()
        self.classification_name = EmptyClass()

        dfs = read_csv_files_as_dataframes(
            directory_path=path_to_resource,
            exclude_file=None,  # ,"resource.csv",
            prefix_dtype_mapping=dtype_dict,
        )

        def _remove_keywords(text, keywords):
            for keyword in keywords:
                text = text.replace(keyword, "")
            return text

        for file_name, df in dfs.items():

            if isinstance(df, pd.DataFrame):  # and "resources.csv" not in file_name:
                setattr(self.datapackage, file_name, df)

        self._create_attributes()
        self._create_attributes_names()

        # Inject the CustomDataFrame class into the DataFrame objects in datapackage
        # Iterate over DataFrame attributes in datapackage
        for attr_name in dir(self.datapackage):
            attr = getattr(self.datapackage, attr_name)
            if isinstance(attr, pd.DataFrame):
                # Replace the DataFrame with an instance of CustomDataFrame
                setattr(self.datapackage, attr_name, CustomDataFrame(attr))

    def _create_attributes(self):
        def _remove_keywords(text, keywords):
            for keyword in keywords:
                text = text.replace(keyword, "")
            return text

        if hasattr(self.datapackage, "tree_bonsai"):
            for index, row in self.datapackage.tree_bonsai.iterrows():
                if "compartment" in self.datapackage.tree_bonsai.columns:
                    obj = TreeMasterFlowobject(
                        code=row["code"],
                        parent_code=row["parent_code"],
                        name=row["name"],
                        level=row["level"],
                        default_unit=row["default_unit"],
                        compartment=row["compartment"],
                        chemical_compound=row["chemical_compound"],
                        alias_code=row["alias_code"],
                    )
                else:
                    obj = BonsaiTreeMaster(
                        code=row["code"],
                        parent_code=row["parent_code"],
                        name=row["name"],
                        level=row["level"],
                        alias_code=row["alias_code"],
                    )
                setattr(self.bonsai, row["code"], obj)
        elif hasattr(self.datapackage, "dim_bonsai"):
            for index, row in self.datapackage.dim_bonsai.iterrows():
                obj = BonsaiDimMaster(
                    code=row["code"],
                    name=row["name"],
                    description=row["description"],
                    comment=row["comment"],
                )
                setattr(self.bonsai, row["code"], obj)

    def _create_attributes_names(self):
        def _remove_keywords(text, keywords):
            for keyword in keywords:
                text = text.replace(keyword, "")
            return text

        if hasattr(self.datapackage, "resources"):
            exclude_strings = [
                "level",
                "compartment",
                "chemical_compound",
                "lcia",
                "urban_rural",
                "distribution",
                "calendar",
                "year",
                "unit",
                "unit_conversion",
            ]
            for index, row in self.datapackage.resources.iterrows():
                _name = row["name"]
                if ("tree_" in _name or "dim_" in _name) and not any(
                    excl in _name for excl in exclude_strings
                ):
                    class_name = _remove_keywords(_name, ["tree_", "dim_"])
                    obj = ObjectNames(code=class_name)
                    # setattr(
                    #    self.classification_name,
                    #    class_name,
                    #    class_name,
                    # )
                    setattr(self.classification_name, class_name, obj)

    def get_classification_names(self):
        """
        Return considerred classification names.

        Returns
        -------
        list of strings
        """
        attribute_list = [
            attr for attr in dir(self.classification_name) if not attr.startswith("__")
        ]
        return attribute_list


def get_bonsai_classification():
    """
    Returns a dictionary of all the default bonsai classification names
    """
    bonsai_classifications = {
        "location": "bonsai",
        "activitytype": "bonsai",
        "flowobject": "bonsai",
        "flow": "bonsai",
    }
    return bonsai_classifications


def get_bonsai_schemas_mapping():
    """
    Returns dictionary that mapps bonsai schemas to bonsai codes.
    """
    mapping = {
        "ProductionVolumes": {"activitytype": ["ai"], "flowobject": ["fi"]},
        "Trade": {"activitytype": ["oa_imp", "oa_exp"], "flowobject": ["fi"]},
        "ConsumptionVolumes": {"activitytype": ["ai"], "flowobject": ["fi"]},
        "Use": {"activitytype": ["ai", "am", "aa_comp"], "flowobject": ["fi", "fm"]},
        "WasteSupply": {"activitytype": ["at"], "flowobject": ["ft"]},
        "Supply": {"activitytype": ["ai", "am", "aa_comp"], "flowobject": ["fi", "fm"]},
        "Imports": {"activitytype": ["oa_imp"], "flowobject": ["fi"]},
        "FinalUses": {"activitytype": ["oa_FU"], "flowobject": ["fi"]},
        "ValueAdded": {"activitytype": ["ai"], "flowobject": ["fec_VA"]},
        "SocialSatellite": {"activitytype": ["ai"], "flowobject": ["fs"]},
        "Valuation": {"activitytype": ["oa_VALU"], "flowobject": ["fi"]},
    }
    return mapping


activitytype = Resources(ROOT_PATH.joinpath(activitytype_path))
location = Resources(ROOT_PATH.joinpath(location_path))
dataquality = Resources(ROOT_PATH.joinpath(dataquality_path))
uncertainty = Resources(ROOT_PATH.joinpath(uncertainty_path))
time = Resources(ROOT_PATH.joinpath(time_path))
flowobject = Resources(ROOT_PATH.joinpath(flowobject_path))
flow = Resources(ROOT_PATH.joinpath(flow_path))
unit_monetary = Resources(ROOT_PATH.joinpath(unit_monetary_path))
unit_physical = Resources(ROOT_PATH.joinpath(unit_physical_path))
unit_magnitude = Resources(ROOT_PATH.joinpath(unit_magnitude_path))
